from unifi_network_maps.model.topology import Edge
from unifi_network_maps.render.mermaid import render_legend, render_legend_compact, render_mermaid


def test_render_mermaid_uses_ids_with_labels():
    output = render_mermaid([Edge("AP Wifi6 tuinhuis", "Core Switch")])
    assert 'ap_wifi6_tuinhuis["AP Wifi6 tuinhuis"]' in output


def test_render_mermaid_includes_edge_label():
    output = render_mermaid([Edge("A", "B", label="Port 1")])
    assert '---|"Port 1"|' in output


def test_render_mermaid_styles_poe_links():
    output = render_mermaid([Edge("A", "B", poe=True)])
    assert "linkStyle 0 stroke:#1e88e5" in output


def test_render_mermaid_styles_wireless_links():
    output = render_mermaid([Edge("A", "B", wireless=True)])
    assert "stroke-dasharray: 5 4" in output


def test_render_legend_outputs_subgraph():
    output = render_legend()
    assert "subgraph legend" in output


def test_render_mermaid_grouped_uses_semicolons():
    output = render_mermaid(
        [Edge("A", "B")],
        groups={"group": ["A", "B"]},
        group_order=["group"],
        node_types={"A": "gateway", "B": "switch"},
    )
    lines = output.splitlines()
    assert '  subgraph group_group["Group"];' in lines
    assert any(line.endswith(";") for line in lines if "---" in line)
    assert any(line.endswith(";") for line in lines if line.strip().startswith("class "))


def test_render_legend_link_inside_subgraph():
    output = render_legend().splitlines()
    link_line = "    legend_poe_a ---|⚡| legend_poe_b;"
    end_line = "  end"
    assert link_line in output
    assert output.index(link_line) < output.index(end_line)
    assert "    legend_no_poe_a --- legend_no_poe_b;" in output
    assert "    linkStyle 0 arrowhead:none;" in output
    assert "    linkStyle 1 arrowhead:none;" in output


def test_render_legend_subgraph_ends_with_semicolon():
    output = render_legend().splitlines()
    assert '  subgraph legend["Legend"];' in output


def test_render_legend_link_style_default():
    output = render_legend().splitlines()
    assert "  linkStyle 0 stroke:#1e88e5,stroke-width:2px,arrowhead:none;" in output
    assert "  linkStyle 1 stroke:#2ecc71,stroke-width:2px,arrowhead:none;" in output


def test_render_legend_class_lines_end_with_semicolon():
    output = render_legend().splitlines()
    assert "  class legend_gateway node_gateway;" in output


def test_render_legend_compact_outputs_table():
    output = render_legend_compact()
    assert '<table class="unifi-legend-table">' in output
    assert "background:#ffe3b3" in output
    assert "Link</span>" in output
    assert "Link</span>" in output and "⚡" not in output.split("Link</span>")[1]


def test_render_mermaid_renders_group_subgraph():
    output = render_mermaid([Edge("Gateway", "Switch")], groups={"gateway": ["Gateway"]})
    assert "subgraph group_gateway" in output


def test_render_mermaid_assigns_class_for_node_types():
    output = render_mermaid([Edge("A", "B")], node_types={"A": "gateway"})
    assert "class a node_gateway" in output


def test_render_mermaid_escapes_quotes():
    output = render_mermaid([Edge('A "1"', "B")])
    assert '\\"' in output


def test_render_mermaid_escapes_backslashes():
    output = render_mermaid([Edge("A \\ 1", "B")])
    assert "\\\\" in output


def test_render_mermaid_escapes_newlines():
    output = render_mermaid([Edge("Line 1\nLine 2", "B")])
    assert "\\n" in output



"""Theme loading for Mermaid and SVG rendering."""

from __future__ import annotations

from pathlib import Path

import yaml

from .mermaid_theme import DEFAULT_THEME as DEFAULT_MERMAID_THEME
from .mermaid_theme import MermaidTheme
from .svg_theme import DEFAULT_THEME as DEFAULT_SVG_THEME
from .svg_theme import SvgTheme


def _coerce_pair(value: object, default: tuple[str, str]) -> tuple[str, str]:
    if isinstance(value, list | tuple) and len(value) == 2:
        left, right = value
        if isinstance(left, str) and isinstance(right, str):
            return (left, right)
    if isinstance(value, dict):
        left = value.get("from") or value.get("start")
        right = value.get("to") or value.get("end")
        if isinstance(left, str) and isinstance(right, str):
            return (left, right)
    return default


def _coerce_color(value: object, default: str) -> str:
    return value if isinstance(value, str) else default


def _coerce_optional_color(value: object, default: str | None) -> str | None:
    return value if isinstance(value, str) else default


def _coerce_optional_int(value: object, default: int | None) -> int | None:
    if isinstance(value, int):
        return value
    if isinstance(value, float):
        return int(value)
    return default


def _mermaid_theme_from_dict(data: dict, base: MermaidTheme) -> MermaidTheme:
    nodes = data.get("nodes", {}) if isinstance(data.get("nodes"), dict) else {}

    def _node(name: str) -> tuple[str, str]:
        return (
            _coerce_color(nodes.get(name, {}).get("fill"), getattr(base, f"node_{name}")[0]),
            _coerce_color(nodes.get(name, {}).get("stroke"), getattr(base, f"node_{name}")[1]),
        )

    return MermaidTheme(
        node_gateway=_node("gateway"),
        node_switch=_node("switch"),
        node_ap=_node("ap"),
        node_client=_node("client"),
        node_other=_node("other"),
        poe_link=_coerce_color(data.get("poe_link"), base.poe_link),
        poe_link_width=int(data.get("poe_link_width", base.poe_link_width)),
        poe_link_arrow=_coerce_color(data.get("poe_link_arrow"), base.poe_link_arrow),
        standard_link=_coerce_color(data.get("standard_link"), base.standard_link),
        standard_link_width=int(data.get("standard_link_width", base.standard_link_width)),
        standard_link_arrow=_coerce_color(
            data.get("standard_link_arrow"), base.standard_link_arrow
        ),
        node_text=_coerce_optional_color(data.get("node_text"), base.node_text),
        edge_label_border=_coerce_optional_color(
            data.get("edge_label_border"), base.edge_label_border
        ),
        edge_label_border_width=_coerce_optional_int(
            data.get("edge_label_border_width"), base.edge_label_border_width
        ),
    )


def _svg_theme_from_dict(data: dict, base: SvgTheme) -> SvgTheme:
    nodes = data.get("nodes", {}) if isinstance(data.get("nodes"), dict) else {}
    links = data.get("links", {}) if isinstance(data.get("links"), dict) else {}

    return SvgTheme(
        link_standard=_coerce_pair(links.get("standard"), base.link_standard),
        link_poe=_coerce_pair(links.get("poe"), base.link_poe),
        node_gateway=_coerce_pair(nodes.get("gateway"), base.node_gateway),
        node_switch=_coerce_pair(nodes.get("switch"), base.node_switch),
        node_ap=_coerce_pair(nodes.get("ap"), base.node_ap),
        node_client=_coerce_pair(nodes.get("client"), base.node_client),
        node_other=_coerce_pair(nodes.get("other"), base.node_other),
    )


def load_theme(path: str | Path) -> tuple[MermaidTheme, SvgTheme]:
    theme_path = Path(path)
    payload = yaml.safe_load(theme_path.read_text(encoding="utf-8"))
    if not isinstance(payload, dict):
        raise ValueError("Theme file must contain a YAML mapping")

    mermaid_data = payload.get("mermaid", {})
    svg_data = payload.get("svg", {})

    mermaid_theme = _mermaid_theme_from_dict(mermaid_data, DEFAULT_MERMAID_THEME)
    svg_theme = _svg_theme_from_dict(svg_data, DEFAULT_SVG_THEME)
    return mermaid_theme, svg_theme


def resolve_themes(theme_file: str | Path | None) -> tuple[MermaidTheme, SvgTheme]:
    if theme_file:
        return load_theme(theme_file)
    return DEFAULT_MERMAID_THEME, DEFAULT_SVG_THEME



"""CLI argument definitions."""

from __future__ import annotations

import argparse


def build_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        description="Generate network maps from UniFi LLDP data, as mermaid or SVG"
    )
    add_source_args(parser.add_argument_group("Source"))
    add_mock_args(parser.add_argument_group("Mock"))
    add_functional_args(parser.add_argument_group("Functional"))
    add_mermaid_args(parser.add_argument_group("Mermaid"))
    add_svg_args(parser.add_argument_group("SVG"))
    add_general_render_args(parser.add_argument_group("Output"))
    add_debug_args(parser.add_argument_group("Debug"))
    return parser


def add_source_args(parser: argparse._ArgumentGroup) -> None:
    parser.add_argument("--site", default=None, help="UniFi site name (overrides UNIFI_SITE)")
    parser.add_argument(
        "--env-file",
        default=None,
        help="Path to .env file (overrides default .env discovery)",
    )
    parser.add_argument(
        "--mock-data",
        default=None,
        help="Path to mock data JSON (skips UniFi API calls)",
    )


def add_mock_args(parser: argparse._ArgumentGroup) -> None:
    parser.add_argument(
        "--generate-mock",
        default=None,
        help="Write mock data JSON to the given path and exit",
    )
    parser.add_argument("--mock-seed", type=int, default=1337, help="Seed for mock generation")
    parser.add_argument(
        "--mock-switches",
        type=int,
        default=1,
        help="Number of switches to generate (default: 1)",
    )
    parser.add_argument(
        "--mock-aps",
        type=int,
        default=2,
        help="Number of access points to generate (default: 2)",
    )
    parser.add_argument(
        "--mock-wired-clients",
        type=int,
        default=2,
        help="Number of wired clients to generate (default: 2)",
    )
    parser.add_argument(
        "--mock-wireless-clients",
        type=int,
        default=2,
        help="Number of wireless clients to generate (default: 2)",
    )


def add_functional_args(parser: argparse._ArgumentGroup) -> None:
    parser.add_argument("--include-ports", action="store_true", help="Include port labels in edges")
    parser.add_argument(
        "--include-clients",
        action="store_true",
        help="Include active clients as leaf nodes",
    )
    parser.add_argument(
        "--client-scope",
        choices=["wired", "wireless", "all"],
        default="wired",
        help="Client types to include (default: wired)",
    )
    parser.add_argument(
        "--only-unifi", action="store_true", help="Only include neighbors that are UniFi devices"
    )
    parser.add_argument(
        "--no-cache",
        action="store_true",
        help="Disable UniFi API cache reads and writes",
    )


def add_mermaid_args(parser: argparse._ArgumentGroup) -> None:
    parser.add_argument("--direction", default="TB", choices=["LR", "TB"], help="Mermaid direction")
    parser.add_argument(
        "--group-by-type",
        action="store_true",
        help="Group nodes by gateway/switch/ap in Mermaid subgraphs",
    )
    parser.add_argument(
        "--legend-scale",
        type=float,
        default=1.0,
        help="Scale legend font/link sizes for Mermaid output (default: 1.0)",
    )
    parser.add_argument(
        "--legend-style",
        default="auto",
        choices=["auto", "compact", "diagram"],
        help="Legend style (auto uses compact for mkdocs, diagram otherwise)",
    )
    parser.add_argument(
        "--legend-only",
        action="store_true",
        help="Render only the legend as a separate Mermaid graph",
    )


def add_svg_args(parser: argparse._ArgumentGroup) -> None:
    parser.add_argument("--svg-width", type=int, default=None, help="SVG width override")
    parser.add_argument("--svg-height", type=int, default=None, help="SVG height override")
    parser.add_argument("--theme-file", default=None, help="Path to theme YAML file")


def add_general_render_args(parser: argparse._ArgumentGroup) -> None:
    parser.add_argument(
        "--format",
        default="mermaid",
        choices=["mermaid", "svg", "svg-iso", "lldp-md", "mkdocs"],
        help="Output format",
    )
    parser.add_argument(
        "--markdown",
        action="store_true",
        help="Wrap output in a Markdown mermaid code fence for notes tools like Obsidian",
    )
    parser.add_argument("--output", default=None, help="Output file path")
    parser.add_argument("--stdout", action="store_true", help="Write output to stdout")
    parser.add_argument(
        "--mkdocs-sidebar-legend",
        action="store_true",
        help="For mkdocs output, write sidebar legend assets next to the output file",
    )
    parser.add_argument(
        "--mkdocs-dual-theme",
        action="store_true",
        help="Render light/dark Mermaid blocks for MkDocs Material theme switching",
    )
    parser.add_argument(
        "--mkdocs-timestamp-zone",
        default="Europe/Amsterdam",
        help="Timezone for mkdocs generated timestamp (use 'off' to disable)",
    )


def add_debug_args(parser: argparse._ArgumentGroup) -> None:
    parser.add_argument(
        "--debug-dump",
        action="store_true",
        help="Dump gateway and sample device data to stderr for debugging",
    )
    parser.add_argument(
        "--debug-sample",
        type=int,
        default=2,
        help="Number of non-gateway devices to include in debug dump (default: 2)",
    )



from unifi_network_maps.model.labels import compose_port_label, order_edge_names


def test_compose_port_label_with_both_sides():
    label = compose_port_label("A", "B", {("A", "B"): "Port 1", ("B", "A"): "Port 2"})
    assert label == "A: Port 1 <-> B: Port 2"


def test_compose_port_label_with_left_only():
    label = compose_port_label("A", "B", {("A", "B"): "Port 1"})
    assert label == "A: Port 1 <-> B: ?"


def test_compose_port_label_with_right_only():
    label = compose_port_label("A", "B", {("B", "A"): "Port 2"})
    assert label == "A: ? <-> B: Port 2"


def test_compose_port_label_with_none():
    label = compose_port_label("A", "B", {})
    assert label is None


def test_order_edge_names_swaps_when_right_label_only():
    ordered = order_edge_names("A", "B", {("B", "A"): "Port 2"}, lambda _name: 0)
    assert ordered == ("B", "A")


def test_order_edge_names_prefers_lower_rank():
    ordered = order_edge_names(
        "Switch",
        "Gateway",
        {("Switch", "Gateway"): "Port 1", ("Gateway", "Switch"): "Port 2"},
        lambda name: 1 if name == "Switch" else 0,
    )
    assert ordered == ("Gateway", "Switch")



"""MkDocs asset output helpers."""

from __future__ import annotations

from pathlib import Path

from ..render.templating import render_template


def write_mkdocs_sidebar_assets(output_path: str) -> None:
    output_dir = Path(output_path).resolve().parent
    assets_dir = output_dir / "assets"
    assets_dir.mkdir(parents=True, exist_ok=True)
    (assets_dir / "legend.js").write_text(
        render_template("mkdocs_legend.js.j2"),
        encoding="utf-8",
    )
    (assets_dir / "legend.css").write_text(
        render_template("mkdocs_legend.css.j2"),
        encoding="utf-8",
    )



import builtins
import json
import sys
import time
from types import SimpleNamespace

import pytest

from unifi_network_maps.adapters import unifi
from unifi_network_maps.adapters.config import Config


def test_fetch_devices_falls_back_on_auth_error(monkeypatch):
    class FakeAuthError(Exception):
        pass

    fake_module = SimpleNamespace(UnifiAuthenticationError=FakeAuthError)
    monkeypatch.setitem(sys.modules, "unifi_controller_api", fake_module)

    def fake_init_controller(config, *, is_udm_pro):
        if is_udm_pro:
            raise FakeAuthError("bad auth")

        class Controller:
            def get_unifi_site_device(self, site_name, detailed, raw):
                return [object(), object()]

        return Controller()

    monkeypatch.setattr(unifi, "_init_controller", fake_init_controller)
    config = Config(
        url="https://example", site="default", user="user", password="pass", verify_ssl=True
    )
    devices = list(unifi.fetch_devices(config))
    assert len(devices) == 2


def test_fetch_devices_requires_dependency(monkeypatch):
    real_import = builtins.__import__

    def fake_import(name, *args, **kwargs):
        if name == "unifi_controller_api":
            raise ImportError("missing")
        return real_import(name, *args, **kwargs)

    monkeypatch.setattr(builtins, "__import__", fake_import)
    config = Config(
        url="https://example", site="default", user="user", password="pass", verify_ssl=True
    )
    with pytest.raises(RuntimeError) as excinfo:
        unifi.fetch_devices(config)
    assert "Missing dependency" in str(excinfo.value)


def test_init_controller_passes_config(monkeypatch):
    captured = {}

    class FakeController:
        def __init__(self, **kwargs):
            captured.update(kwargs)

    fake_module = SimpleNamespace(UnifiController=FakeController)
    monkeypatch.setitem(sys.modules, "unifi_controller_api", fake_module)
    config = Config(
        url="https://example", site="default", user="user", password="pass", verify_ssl=False
    )
    unifi._init_controller(config, is_udm_pro=True)
    assert captured["verify_ssl"] is False


def test_fetch_clients_falls_back_on_auth_error(monkeypatch):
    class FakeAuthError(Exception):
        pass

    fake_module = SimpleNamespace(UnifiAuthenticationError=FakeAuthError)
    monkeypatch.setitem(sys.modules, "unifi_controller_api", fake_module)

    def fake_init_controller(config, *, is_udm_pro):
        if is_udm_pro:
            raise FakeAuthError("bad auth")

        class Controller:
            def get_unifi_site_client(self, site_name, raw):
                return [object()]

        return Controller()

    monkeypatch.setattr(unifi, "_init_controller", fake_init_controller)
    config = Config(
        url="https://example", site="default", user="user", password="pass", verify_ssl=True
    )
    clients = list(unifi.fetch_clients(config))
    assert len(clients) == 1


def test_fetch_clients_requires_dependency(monkeypatch):
    real_import = builtins.__import__

    def fake_import(name, *args, **kwargs):
        if name == "unifi_controller_api":
            raise ImportError("missing")
        return real_import(name, *args, **kwargs)

    monkeypatch.setattr(builtins, "__import__", fake_import)
    config = Config(
        url="https://example", site="default", user="user", password="pass", verify_ssl=True
    )
    with pytest.raises(RuntimeError) as excinfo:
        unifi.fetch_clients(config)
    assert "Missing dependency" in str(excinfo.value)


def test_fetch_devices_uses_cache(monkeypatch, tmp_path):
    fake_module = SimpleNamespace(UnifiAuthenticationError=RuntimeError)
    monkeypatch.setitem(sys.modules, "unifi_controller_api", fake_module)
    monkeypatch.setenv("UNIFI_CACHE_DIR", str(tmp_path))
    monkeypatch.setenv("UNIFI_CACHE_TTL_SECONDS", "3600")

    config = Config(
        url="https://example", site="default", user="user", password="pass", verify_ssl=True
    )
    cache_path = tmp_path / f"devices_{unifi._cache_key(config.url, config.site, 'True')}.json"
    unifi._save_cache(cache_path, [{"name": "cached"}])

    def fail_init(*_args, **_kwargs):
        raise AssertionError("should not fetch when cache is valid")

    monkeypatch.setattr(unifi, "_init_controller", fail_init)
    devices = list(unifi.fetch_devices(config))
    device = devices[0]
    assert isinstance(device, dict)
    assert device["name"] == "cached"


def test_fetch_devices_skips_cache_when_disabled(monkeypatch, tmp_path):
    fake_module = SimpleNamespace(UnifiAuthenticationError=RuntimeError)
    monkeypatch.setitem(sys.modules, "unifi_controller_api", fake_module)
    monkeypatch.setenv("UNIFI_CACHE_DIR", str(tmp_path))
    monkeypatch.setenv("UNIFI_CACHE_TTL_SECONDS", "3600")

    cache_path = tmp_path / f"devices_{unifi._cache_key('url', 'default', 'True')}.json"
    cache_path.write_text(
        json.dumps({"timestamp": time.time(), "data": [{"name": "cached"}]}),
        encoding="utf-8",
    )

    calls = {"count": 0}

    class Controller:
        def get_unifi_site_device(self, site_name, detailed, raw):
            calls["count"] += 1
            return [{"name": "fresh"}]

    monkeypatch.setattr(unifi, "_init_controller", lambda *_a, **_k: Controller())
    config = Config(url="url", site="default", user="user", password="pass", verify_ssl=True)
    devices = list(unifi.fetch_devices(config, use_cache=False))
    device = devices[0]
    assert calls["count"] == 1
    assert isinstance(device, dict)
    assert device["name"] == "fresh"


def test_fetch_clients_cache_expired(monkeypatch, tmp_path):
    fake_module = SimpleNamespace(UnifiAuthenticationError=RuntimeError)
    monkeypatch.setitem(sys.modules, "unifi_controller_api", fake_module)
    monkeypatch.setenv("UNIFI_CACHE_DIR", str(tmp_path))
    monkeypatch.setenv("UNIFI_CACHE_TTL_SECONDS", "1")

    config = Config(
        url="https://example", site="default", user="user", password="pass", verify_ssl=True
    )
    cache_path = tmp_path / f"clients_{unifi._cache_key(config.url, config.site)}.json"
    cache_path.write_text(
        json.dumps({"timestamp": time.time() - 3600, "data": [{"stale": True}]}),
        encoding="utf-8",
    )

    class Controller:
        def get_unifi_site_client(self, site_name, raw):
            return [{"fresh": True}]

    monkeypatch.setattr(unifi, "_init_controller", lambda *_a, **_k: Controller())
    clients = list(unifi.fetch_clients(config))
    client = clients[0]
    assert isinstance(client, dict)
    assert client["fresh"] is True


def test_fetch_devices_uses_stale_cache_on_error(monkeypatch, tmp_path):
    fake_module = SimpleNamespace(UnifiAuthenticationError=RuntimeError)
    monkeypatch.setitem(sys.modules, "unifi_controller_api", fake_module)
    monkeypatch.setenv("UNIFI_CACHE_DIR", str(tmp_path))
    monkeypatch.setenv("UNIFI_CACHE_TTL_SECONDS", "1")

    config = Config(
        url="https://example", site="default", user="user", password="pass", verify_ssl=True
    )
    cache_path = tmp_path / f"devices_{unifi._cache_key(config.url, config.site, 'True')}.json"
    cache_path.write_text(
        json.dumps({"timestamp": time.time() - 3600, "data": [{"stale": True}]}),
        encoding="utf-8",
    )

    class Controller:
        def get_unifi_site_device(self, site_name, detailed, raw):
            raise RuntimeError("boom")

    monkeypatch.setattr(unifi, "_init_controller", lambda *_a, **_k: Controller())
    devices = list(unifi.fetch_devices(config))
    device = devices[0]
    assert isinstance(device, dict)
    assert device["stale"] is True


def test_fetch_clients_uses_stale_cache_on_error(monkeypatch, tmp_path):
    fake_module = SimpleNamespace(UnifiAuthenticationError=RuntimeError)
    monkeypatch.setitem(sys.modules, "unifi_controller_api", fake_module)
    monkeypatch.setenv("UNIFI_CACHE_DIR", str(tmp_path))
    monkeypatch.setenv("UNIFI_CACHE_TTL_SECONDS", "1")

    config = Config(
        url="https://example", site="default", user="user", password="pass", verify_ssl=True
    )
    cache_path = tmp_path / f"clients_{unifi._cache_key(config.url, config.site)}.json"
    cache_path.write_text(
        json.dumps({"timestamp": time.time() - 3600, "data": [{"stale": True}]}),
        encoding="utf-8",
    )

    class Controller:
        def get_unifi_site_client(self, site_name, raw):
            raise RuntimeError("boom")

    monkeypatch.setattr(unifi, "_init_controller", lambda *_a, **_k: Controller())
    clients = list(unifi.fetch_clients(config))
    client = clients[0]
    assert isinstance(client, dict)
    assert client["stale"] is True


def test_fetch_devices_retries(monkeypatch, tmp_path):
    fake_module = SimpleNamespace(UnifiAuthenticationError=RuntimeError)
    monkeypatch.setitem(sys.modules, "unifi_controller_api", fake_module)
    monkeypatch.setenv("UNIFI_RETRY_ATTEMPTS", "2")
    monkeypatch.setenv("UNIFI_RETRY_BACKOFF_SECONDS", "0")
    monkeypatch.setenv("UNIFI_CACHE_TTL_SECONDS", "0")
    monkeypatch.setenv("UNIFI_CACHE_DIR", str(tmp_path))

    calls = {"count": 0}

    class Controller:
        def get_unifi_site_device(self, site_name, detailed, raw):
            calls["count"] += 1
            if calls["count"] == 1:
                raise RuntimeError("boom")
            return [{"ok": True}]

    monkeypatch.setattr(unifi, "_init_controller", lambda *_a, **_k: Controller())
    config = Config(
        url="https://example", site="default", user="user", password="pass", verify_ssl=True
    )
    devices = list(unifi.fetch_devices(config))
    assert calls["count"] == 2
    device = devices[0]
    assert isinstance(device, dict)
    assert device["ok"] is True


def test_call_with_retries_times_out(monkeypatch):
    monkeypatch.setenv("UNIFI_RETRY_ATTEMPTS", "1")
    monkeypatch.setenv("UNIFI_RETRY_BACKOFF_SECONDS", "0")
    monkeypatch.setenv("UNIFI_REQUEST_TIMEOUT_SECONDS", "0.01")

    def slow_call():
        time.sleep(0.05)
        return "ok"

    with pytest.raises(TimeoutError):
        unifi._call_with_retries("slow", slow_call)


def test_fetch_devices_skips_cache_when_dir_is_world_writable(monkeypatch, tmp_path):
    fake_module = SimpleNamespace(UnifiAuthenticationError=RuntimeError)
    monkeypatch.setitem(sys.modules, "unifi_controller_api", fake_module)
    monkeypatch.setenv("UNIFI_CACHE_DIR", str(tmp_path))
    monkeypatch.setenv("UNIFI_CACHE_TTL_SECONDS", "3600")

    cache_path = tmp_path / f"devices_{unifi._cache_key('url', 'default', 'True')}.json"
    cache_path.parent.mkdir(parents=True, exist_ok=True)
    cache_path.write_text(
        json.dumps({"timestamp": time.time(), "data": [{"name": "cached"}]}),
        encoding="utf-8",
    )
    tmp_path.chmod(0o777)

    called = {"count": 0}

    class Controller:
        def get_unifi_site_device(self, site_name, detailed, raw):
            called["count"] += 1
            return [{"name": "fresh"}]

    monkeypatch.setattr(unifi, "_init_controller", lambda *_a, **_k: Controller())
    config = Config(url="url", site="default", user="user", password="pass", verify_ssl=True)
    devices = list(unifi.fetch_devices(config))
    device = devices[0]
    assert called["count"] == 1
    assert isinstance(device, dict)
    assert device["name"] == "fresh"



"""Generate mock UniFi data using Faker."""

from __future__ import annotations

import json
import random
from dataclasses import dataclass, field
from typing import Any

from faker import Faker


@dataclass(frozen=True)
class MockOptions:
    seed: int = 1337
    switch_count: int = 1
    ap_count: int = 2
    wired_client_count: int = 2
    wireless_client_count: int = 2


@dataclass
class _MockState:
    fake: Faker
    rng: random.Random
    used_macs: set[str] = field(default_factory=set)
    used_ips: set[str] = field(default_factory=set)
    used_names: set[str] = field(default_factory=set)
    used_rooms: set[str] = field(default_factory=set)
    core_port_next: int = 2


def generate_mock_payload(options: MockOptions) -> dict[str, list[dict[str, Any]]]:
    state = _build_state(options.seed)
    devices, core_switch, aps = _build_devices(options, state)
    clients = _build_clients(options, state, core_switch, aps)
    return {"devices": devices, "clients": clients}


def mock_payload_json(options: MockOptions) -> str:
    payload = generate_mock_payload(options)
    return json.dumps(payload, indent=2, sort_keys=True)


def _build_state(seed: int) -> _MockState:
    fake = Faker("en_US")
    fake.seed_instance(seed)
    rng = random.Random(seed)
    return _MockState(fake=fake, rng=rng)


def _build_devices(
    options: MockOptions, state: _MockState
) -> tuple[list[dict[str, Any]], dict[str, Any], list[dict[str, Any]]]:
    gateway = _build_gateway(state)
    core_switch = _build_core_switch(state)
    _link_gateway_to_switch(state, gateway, core_switch)
    access_switches = _build_access_switches(options.switch_count - 1, state, core_switch)
    aps = _build_aps(options.ap_count, state, core_switch)
    devices = [gateway, core_switch] + access_switches + aps
    return devices, core_switch, aps


def _build_gateway(state: _MockState) -> dict[str, Any]:
    device = _device_base(state, "Cloud Gateway", "udm", "UniFi Dream Machine Pro", "UDM-Pro")
    _add_port(device, 9, poe_enabled=False, rng=state.rng)
    return device


def _build_core_switch(state: _MockState) -> dict[str, Any]:
    return _device_base(state, "Core Switch", "usw", "UniFi Switch 24 PoE", "USW-24-PoE")


def _build_access_switches(
    count: int, state: _MockState, core_switch: dict[str, Any]
) -> list[dict[str, Any]]:
    switches = []
    for _ in range(max(0, count)):
        room = _unique_room(state)
        name = _unique_name(state, f"Switch {room}")
        device = _device_base(state, name, "usw", "UniFi Switch Lite 8 PoE", "USW-Lite-8-PoE")
        _link_core_device(state, core_switch, device, poe_enabled=False)
        switches.append(device)
    return switches


def _build_aps(count: int, state: _MockState, core_switch: dict[str, Any]) -> list[dict[str, Any]]:
    aps = []
    for _ in range(max(0, count)):
        room = _unique_room(state)
        name = _unique_name(state, f"AP {room}")
        device = _device_base(state, name, "uap", "UniFi AP 6 Lite", "U6-Lite")
        _add_port(device, 1, poe_enabled=True, rng=state.rng)
        _link_core_device(state, core_switch, device, poe_enabled=True)
        aps.append(device)
    return aps


def _build_clients(
    options: MockOptions,
    state: _MockState,
    core_switch: dict[str, Any],
    aps: list[dict[str, Any]],
) -> list[dict[str, Any]]:
    clients = []
    clients.extend(_build_wired_clients(options.wired_client_count, state, core_switch))
    clients.extend(_build_wireless_clients(options.wireless_client_count, state, aps))
    return clients


def _build_wired_clients(
    count: int, state: _MockState, core_switch: dict[str, Any]
) -> list[dict[str, Any]]:
    clients = []
    for _ in range(max(0, count)):
        port_idx = _next_core_port(state)
        _add_port(core_switch, port_idx, poe_enabled=False, rng=state.rng)
        name = _unique_client_name(state)
        clients.append(
            {
                "name": name,
                "is_wired": True,
                "sw_mac": core_switch["mac"],
                "sw_port": port_idx,
            }
        )
    return clients


def _build_wireless_clients(
    count: int, state: _MockState, aps: list[dict[str, Any]]
) -> list[dict[str, Any]]:
    if not aps:
        return []
    clients = []
    for idx in range(max(0, count)):
        ap = aps[idx % len(aps)]
        name = _unique_client_name(state)
        clients.append(
            {
                "name": name,
                "is_wired": False,
                "ap_mac": ap["mac"],
                "ap_port": 1,
            }
        )
    return clients


def _device_base(
    state: _MockState, name: str, dev_type: str, model_name: str, model: str
) -> dict[str, Any]:
    version = _pick_version(state, dev_type)
    return {
        "name": name,
        "model_name": model_name,
        "model": model,
        "mac": _unique_mac(state),
        "ip": _unique_ip(state),
        "type": dev_type,
        "version": version,
        "port_table": [],
        "lldp_info": [],
    }


def _pick_version(state: _MockState, dev_type: str) -> str:
    versions = {
        "udm": ["3.1.0", "3.1.1"],
        "usw": ["7.0.0", "7.1.2"],
        "uap": ["6.6.55", "6.7.10"],
    }
    return state.rng.choice(versions.get(dev_type, ["1.0.0"]))


def _link_gateway_to_switch(
    state: _MockState, gateway: dict[str, Any], core_switch: dict[str, Any]
) -> None:
    _add_port(core_switch, 1, poe_enabled=False, rng=state.rng)
    _add_lldp_link(
        gateway,
        core_switch,
        local_port=9,
        remote_port=1,
        remote_name=core_switch["name"],
    )
    _add_lldp_link(
        core_switch,
        gateway,
        local_port=1,
        remote_port=9,
        remote_name=gateway["name"],
    )


def _link_core_device(
    state: _MockState,
    core_switch: dict[str, Any],
    device: dict[str, Any],
    *,
    poe_enabled: bool,
) -> None:
    port_idx = _next_core_port(state)
    _add_port(core_switch, port_idx, poe_enabled=poe_enabled, rng=state.rng)
    _add_lldp_link(
        core_switch,
        device,
        local_port=port_idx,
        remote_port=1,
        remote_name=device["name"],
    )
    _add_lldp_link(
        device,
        core_switch,
        local_port=1,
        remote_port=port_idx,
        remote_name=core_switch["name"],
    )


def _add_lldp_link(
    source: dict[str, Any],
    target: dict[str, Any],
    *,
    local_port: int,
    remote_port: int,
    remote_name: str,
) -> None:
    source["lldp_info"].append(
        {
            "chassis_id": target["mac"],
            "port_id": f"Port {remote_port}",
            "port_desc": remote_name,
            "local_port_idx": local_port,
            "local_port_name": f"Port {local_port}",
        }
    )


def _add_port(
    device: dict[str, Any],
    port_idx: int,
    *,
    poe_enabled: bool,
    rng: random.Random,
) -> None:
    device["port_table"].append(
        {
            "port_idx": port_idx,
            "name": f"Port {port_idx}",
            "ifname": f"eth{port_idx}",
            "is_uplink": False,
            "poe_enable": poe_enabled,
            "port_poe": poe_enabled,
            "poe_class": 4 if poe_enabled else 0,
            "poe_power": round(rng.uniform(2.0, 6.0), 2) if poe_enabled else 0.0,
            "poe_good": poe_enabled,
            "poe_voltage": round(rng.uniform(44.0, 52.0), 1) if poe_enabled else 0.0,
            "poe_current": round(rng.uniform(0.05, 0.12), 3) if poe_enabled else 0.0,
        }
    )


def _next_core_port(state: _MockState) -> int:
    port_idx = state.core_port_next
    state.core_port_next += 1
    return port_idx


def _unique_name(state: _MockState, prefix: str) -> str:
    name = prefix
    while name in state.used_names:
        name = f"{prefix} {state.rng.randint(2, 9)}"
    state.used_names.add(name)
    return name


def _unique_client_name(state: _MockState) -> str:
    name = state.fake.first_name()
    while name in state.used_names:
        name = state.fake.first_name()
    state.used_names.add(name)
    return name


def _unique_room(state: _MockState) -> str:
    room = state.fake.word().title()
    while room in state.used_rooms:
        room = state.fake.word().title()
    state.used_rooms.add(room)
    return room


def _unique_mac(state: _MockState) -> str:
    mac = state.fake.mac_address()
    while mac in state.used_macs:
        mac = state.fake.mac_address()
    state.used_macs.add(mac)
    return mac


def _unique_ip(state: _MockState) -> str:
    ip = state.fake.ipv4_private()
    while ip in state.used_ips:
        ip = state.fake.ipv4_private()
    state.used_ips.add(ip)
    return ip



"""Legend rendering helpers."""

from __future__ import annotations

from .mermaid import render_legend, render_legend_compact
from .mermaid_theme import MermaidTheme


def resolve_legend_style(*, format_name: str, legend_style: str) -> str:
    if legend_style == "auto":
        return "compact" if format_name == "mkdocs" else "diagram"
    return legend_style


def render_legend_only(
    *,
    legend_style: str,
    legend_scale: float,
    markdown: bool,
    theme: MermaidTheme,
) -> str:
    if legend_style == "compact":
        content = "# Legend\n\n" + render_legend_compact(theme=theme)
    else:
        content = render_legend(theme=theme, legend_scale=legend_scale)
    if markdown:
        content = f"""```mermaid
{content}```
"""
    return content



"""Generate mock UniFi data using Faker."""

from __future__ import annotations

from ..model.mock import MockOptions, generate_mock_payload, mock_payload_json

__all__ = ["MockOptions", "generate_mock_payload", "mock_payload_json"]



"""Output helpers for files and stdout."""

from __future__ import annotations

import os
import sys
import tempfile
from pathlib import Path


def write_output(content: str, *, output_path: str | None, stdout: bool) -> None:
    if output_path:
        _write_atomic(Path(output_path), content)
    if stdout or not output_path:
        sys.stdout.write(content)


def _write_atomic(path: Path, content: str) -> None:
    path.parent.mkdir(parents=True, exist_ok=True)
    with tempfile.NamedTemporaryFile(
        mode="w",
        encoding="utf-8",
        dir=str(path.parent),
        prefix=f".{path.name}.",
        suffix=".tmp",
        delete=False,
    ) as temp_file:
        temp_file.write(content)
        temp_file.flush()
        os.fsync(temp_file.fileno())
        tmp_path = Path(temp_file.name)
    tmp_path.replace(path)



import pytest

from unifi_network_maps.adapters.config import Config, _parse_bool


def test_parse_bool_true_default():
    assert _parse_bool(None, default=True) is True


def test_parse_bool_false():
    assert _parse_bool("false", default=True) is False


def test_parse_bool_unknown_uses_default():
    assert _parse_bool("maybe", default=False) is False


def test_config_from_env_missing_url(monkeypatch):
    monkeypatch.setenv("UNIFI_URL", "")
    monkeypatch.setenv("UNIFI_USER", "user")
    monkeypatch.setenv("UNIFI_PASS", "pass")
    with pytest.raises(ValueError) as excinfo:
        Config.from_env()
    assert "UNIFI_URL is required" in str(excinfo.value)


def test_config_from_env_missing_user(monkeypatch):
    monkeypatch.setenv("UNIFI_URL", "https://example.local")
    monkeypatch.setenv("UNIFI_USER", "")
    monkeypatch.setenv("UNIFI_PASS", "pass")
    with pytest.raises(ValueError) as excinfo:
        Config.from_env()
    assert "UNIFI_USER is required" in str(excinfo.value)


def test_config_from_env_missing_password(monkeypatch):
    monkeypatch.setenv("UNIFI_URL", "https://example.local")
    monkeypatch.setenv("UNIFI_USER", "user")
    monkeypatch.setenv("UNIFI_PASS", "")
    with pytest.raises(ValueError) as excinfo:
        Config.from_env()
    assert "UNIFI_PASS is required" in str(excinfo.value)


def test_config_from_env_success(monkeypatch):
    monkeypatch.setenv("UNIFI_URL", "https://example.local")
    monkeypatch.setenv("UNIFI_SITE", "default")
    monkeypatch.setenv("UNIFI_USER", "user")
    monkeypatch.setenv("UNIFI_PASS", "pass")
    monkeypatch.setenv("UNIFI_VERIFY_SSL", "false")
    config = Config.from_env()
    assert config.verify_ssl is False


def test_config_from_env_requires_dotenv_for_env_file(monkeypatch):
    import builtins

    real_import = builtins.__import__

    def fake_import(name, *args, **kwargs):
        if name == "dotenv":
            raise ImportError("missing")
        return real_import(name, *args, **kwargs)

    monkeypatch.setattr(builtins, "__import__", fake_import)
    with pytest.raises(ValueError, match="python-dotenv required"):
        Config.from_env(env_file="custom.env")



"""Render per-device port overview tables."""

from __future__ import annotations

from collections import defaultdict
from html import escape as _escape_html

from ..model.ports import extract_port_number
from ..model.topology import ClientPortMap, Device, PortInfo, PortMap, classify_device_type
from .markdown_tables import markdown_table_lines
from .templating import render_template


def render_device_port_overview(
    devices: list[Device],
    port_map: PortMap,
    *,
    client_ports: ClientPortMap | None = None,
) -> str:
    gateways = _collect_devices_by_type(devices, "gateway")
    switches = _collect_devices_by_type(devices, "switch")
    sections: list[str] = []
    if gateways:
        sections.append(
            render_template(
                "markdown_section.md.j2",
                title="Gateways",
                body=_render_device_group(gateways, port_map, client_ports),
            ).rstrip()
        )
    if switches:
        sections.append(
            render_template(
                "markdown_section.md.j2",
                title="Switches",
                body=_render_device_group(switches, port_map, client_ports),
            ).rstrip()
        )
    return "\n\n".join(section for section in sections if section).rstrip() + "\n"


def _collect_devices_by_type(devices: list[Device], desired_type: str) -> list[Device]:
    return sorted(
        [device for device in devices if classify_device_type(device) == desired_type],
        key=lambda item: item.name.lower(),
    )


def _render_device_group(
    devices: list[Device],
    port_map: PortMap,
    client_ports: ClientPortMap | None,
) -> str:
    blocks: list[str] = []
    for device in devices:
        blocks.append(
            render_template(
                "device_port_block.md.j2",
                device_name=device.name,
                details="\n".join(_render_device_details(device)).rstrip(),
                ports="\n".join(_render_device_ports(device, port_map, client_ports)).rstrip(),
            ).rstrip()
        )
    return "\n\n".join(block for block in blocks if block)


def render_device_port_details(
    device: Device,
    port_map: PortMap,
    *,
    client_ports: ClientPortMap | None = None,
) -> str:
    lines = _render_device_details(device)
    lines.extend(_render_device_ports(device, port_map, client_ports))
    return "\n".join(lines).rstrip() + "\n"


def _render_device_ports(
    device: Device,
    port_map: PortMap,
    client_ports: ClientPortMap | None,
) -> list[str]:
    rows = _build_port_rows(device, port_map, client_ports)
    table_rows = [
        [
            _escape_cell(port_label),
            _escape_cell(connected or "-"),
            _escape_cell(speed),
            _escape_cell(poe_state),
            _escape_cell(power),
        ]
        for port_label, connected, speed, poe_state, power in rows
    ]
    lines = ["#### Ports", ""]
    lines.extend(
        markdown_table_lines(
            ["Port", "Connected", "Speed", "PoE", "Power"],
            table_rows,
        )
    )
    return lines


def _build_port_rows(
    device: Device,
    port_map: PortMap,
    client_ports: ClientPortMap | None,
) -> list[tuple[str, str, str, str, str]]:
    connections = _device_port_connections(device.name, port_map)
    client_connections = _device_client_connections(device.name, client_ports)
    aggregated = _aggregate_ports(device.port_table)
    aggregated_indices = {
        port.port_idx
        for ports in aggregated.values()
        for port in ports
        if getattr(port, "port_idx", None) is not None
    }
    rows: list[tuple[tuple[int, int], tuple[str, str, str, str, str]]] = []
    seen_ports: set[int] = set()
    for port in sorted(device.port_table, key=_port_sort_key):
        if port.port_idx in aggregated_indices:
            port_idx = _port_index(port.port_idx, port.name)
            if port_idx is not None:
                seen_ports.add(port_idx)
            continue
        port_idx = _port_index(port.port_idx, port.name)
        if port_idx is not None:
            seen_ports.add(port_idx)
        port_label = _format_port_label(port_idx, port.name)
        connected = _format_connections(
            device.name,
            port_idx,
            connections,
            client_connections,
            port_map,
        )
        rows.append(
            (
                (0, port_idx or 10_000),
                (
                    port_label,
                    connected,
                    _format_speed(port.speed),
                    _format_poe_state(port),
                    _format_poe_power(port.poe_power),
                ),
            )
        )
    for _group_id, group_ports in aggregated.items():
        group_label = _format_aggregate_label(group_ports)
        group_sort = _aggregate_sort_key(group_ports)
        group_connections = _format_aggregate_connections(
            device.name,
            group_ports,
            connections,
            client_connections,
            port_map,
        )
        rows.append(
            (
                (0, group_sort),
                (
                    group_label,
                    group_connections,
                    _format_aggregate_speed(group_ports),
                    _format_aggregate_poe_state(group_ports),
                    _format_aggregate_power(group_ports),
                ),
            )
        )
    for port_idx in sorted(connections):
        if port_idx in seen_ports:
            continue
        port_label = _format_port_label(port_idx, None)
        connected = _format_connections(
            device.name,
            port_idx,
            connections,
            client_connections,
            port_map,
        )
        rows.append(((2, port_idx), (port_label, connected, "-", "-", "-")))
    return [row for _key, row in sorted(rows, key=lambda item: item[0])]


def _device_port_connections(device_name: str, port_map: PortMap) -> dict[int, list[str]]:
    connections: dict[int, list[str]] = defaultdict(list)
    for (src, dst), label in port_map.items():
        if src != device_name:
            continue
        port_idx = extract_port_number(label or "")
        if port_idx is None:
            continue
        connections[port_idx].append(dst)
    return connections


def _device_client_connections(
    device_name: str, client_ports: ClientPortMap | None
) -> dict[int, list[str]]:
    if not client_ports:
        return {}
    rows = client_ports.get(device_name, [])
    connections: dict[int, list[str]] = defaultdict(list)
    for port_idx, name in rows:
        connections[port_idx].append(name)
    return connections


def _format_connections(
    device_name: str,
    port_idx: int | None,
    connections: dict[int, list[str]],
    client_connections: dict[int, list[str]],
    port_map: PortMap,
) -> str:
    if port_idx is None:
        return ""
    peers = connections.get(port_idx, [])
    clients = client_connections.get(port_idx, [])
    if not peers and not clients:
        return ""
    peer_entries: list[str] = []
    for peer in sorted(peers, key=str.lower):
        peer_label = port_map.get((peer, device_name))
        if peer_label:
            peer_entries.append(f"{peer} ({peer_label})")
        else:
            peer_entries.append(peer)
    peer_text = ", ".join(peer_entries)
    client_text = _format_client_connections(clients)
    if peer_text and client_text:
        return f"{peer_text}<br/>{client_text}"
    return peer_text or client_text


def _format_port_label(port_idx: int | None, name: str | None) -> str:
    if name and name.strip():
        normalized = name.strip()
        if port_idx is None:
            return normalized
        if normalized.lower() != f"port {port_idx}".lower():
            return normalized
    if port_idx is None:
        return "Port ?"
    return f"Port {port_idx}"


def _format_speed(speed: int | None) -> str:
    if speed is None or speed <= 0:
        return "-"
    if speed >= 1000:
        if speed % 1000 == 0:
            return f"{speed // 1000}G"
        return f"{speed / 1000:.1f}G"
    return f"{speed}M"


def _format_poe_state(port: object) -> str:
    poe_power = getattr(port, "poe_power", None)
    poe_good = getattr(port, "poe_good", False)
    poe_enable = getattr(port, "poe_enable", False)
    port_poe = getattr(port, "port_poe", False)
    if (poe_power or 0.0) > 0 or poe_good:
        return "active"
    if port_poe or poe_enable:
        if not poe_enable:
            return "disabled"
        return "capable"
    return "-"


def _format_poe_power(power: float | None) -> str:
    if power is None or power <= 0:
        return "-"
    return f"{power:.2f}W"


def _port_index(port_idx: int | None, name: str | None) -> int | None:
    if port_idx is not None:
        return port_idx
    if name:
        return extract_port_number(name)
    return None


def _port_sort_key(port: object) -> tuple[int, str]:
    port_idx = _port_index(getattr(port, "port_idx", None), getattr(port, "name", None))
    if port_idx is not None:
        return (0, f"{port_idx:04d}")
    name = getattr(port, "name", "") or ""
    return (1, name.lower())


def _escape_cell(value: str) -> str:
    return value.replace("|", "\\|")


def _render_device_details(device: Device) -> list[str]:
    lines = [
        "#### Details",
        "",
        "| Field | Value |",
        "| --- | --- |",
        f"| Model | {_escape_cell(_device_model_label(device))} |",
        f"| Type | {_escape_cell(device.type or '-')} |",
        f"| IP | {_escape_cell(device.ip or '-')} |",
        f"| MAC | {_escape_cell(device.mac or '-')} |",
        f"| Firmware | {_escape_cell(device.version or '-')} |",
        f"| Uplink | {_escape_cell(_uplink_summary(device))} |",
        f"| Ports | {_escape_cell(_port_summary(device))} |",
        f"| PoE | {_escape_cell(_poe_summary(device))} |",
        "",
    ]
    return lines


def _port_summary(device: Device) -> str:
    ports = [port for port in device.port_table if port.port_idx is not None]
    if not ports:
        return "-"
    total_ports = len(ports)
    active_ports = sum(1 for port in ports if (port.speed or 0) > 0)
    return f"{total_ports} total, {active_ports} active"


def _poe_summary(device: Device) -> str:
    ports = [port for port in device.port_table if port.port_idx is not None]
    if not ports:
        return "-"
    poe_capable = sum(1 for port in ports if port.port_poe or port.poe_enable)
    poe_active = sum(1 for port in ports if _format_poe_state(port) == "active")
    total_power = sum(port.poe_power or 0.0 for port in ports)
    summary = f"{poe_capable} capable, {poe_active} active"
    if total_power > 0:
        summary = f"{summary}, {total_power:.2f}W"
    return summary


def _uplink_summary(device: Device) -> str:
    uplink = device.uplink or device.last_uplink
    if not uplink:
        if classify_device_type(device) == "gateway":
            return "Internet"
        return "-"
    name = uplink.name or uplink.mac or "Unknown"
    if classify_device_type(device) == "gateway":
        lowered = name.lower()
        if lowered in {"unknown", "wan", "internet"}:
            name = "Internet"
        elif lowered.startswith(("eth", "wan")):
            name = "Internet"
    if uplink.port is not None:
        return f"{name} (Port {uplink.port})"
    return name


def _device_model_label(device: Device) -> str:
    if device.model_name:
        return device.model_name
    if device.model:
        return device.model
    return device.type or "-"


def _format_client_connections(clients: list[str]) -> str:
    if not clients:
        return ""
    if len(clients) == 1:
        return f"{clients[0]} (client)"
    items = "".join(f"<li>{_escape_html(name)}</li>" for name in clients)
    return f'<ul class="unifi-port-clients">{items}</ul>'


def _aggregate_base_groups(port_table: list[PortInfo]) -> dict[str, list[PortInfo]]:
    groups: dict[str, list[PortInfo]] = defaultdict(list)
    for port in port_table:
        group = getattr(port, "aggregation_group", None)
        if group:
            groups[str(group)].append(port)
            continue
        if _looks_like_lag(port):
            port_idx = getattr(port, "port_idx", None)
            if port_idx is not None:
                groups[f"lag-{port_idx}"].append(port)
    return groups


def _extend_singleton_groups(
    groups: dict[str, list[PortInfo]],
    port_table: list[PortInfo],
) -> None:
    if not groups:
        return
    port_by_idx: dict[int, PortInfo] = {
        port.port_idx: port for port in port_table if port.port_idx is not None
    }
    for group_id, group_ports in list(groups.items()):
        if len(group_ports) > 1:
            continue
        lone_port = group_ports[0]
        if not _looks_like_lag(lone_port):
            continue
        port_idx = lone_port.port_idx
        if port_idx is None:
            continue
        candidates: list[PortInfo] = []
        for neighbor in (port_idx - 1, port_idx + 1):
            port = port_by_idx.get(neighbor)
            if port and not getattr(port, "aggregation_group", None):
                if getattr(port, "speed", None) == getattr(lone_port, "speed", None):
                    candidates.append(port)
        if candidates:
            groups[group_id].extend(candidates)


def _aggregate_ports(port_table: list[PortInfo]) -> dict[str, list[PortInfo]]:
    groups = _aggregate_base_groups(port_table)
    _extend_singleton_groups(groups, port_table)
    return groups


def _looks_like_lag(port: PortInfo) -> bool:
    name = (getattr(port, "name", "") or "").lower()
    ifname = (getattr(port, "ifname", "") or "").lower()
    return "lag" in name or "lag" in ifname or "aggregate" in name


def _format_aggregate_label(group_ports: list[PortInfo]) -> str:
    ports = sorted([int(p.port_idx) for p in group_ports if p.port_idx is not None])
    if ports:
        if len(ports) == 1:
            return f"Port {ports[0]} (LAG)"
        if ports == list(range(ports[0], ports[-1] + 1)):
            return f"Port {ports[0]}-{ports[-1]} (LAG)"
        return "Ports " + "+".join(str(port) for port in ports) + " (LAG)"
    return "Aggregated ports"


def _aggregate_sort_key(group_ports: list[PortInfo]) -> int:
    ports = sorted([int(p.port_idx) for p in group_ports if p.port_idx is not None])
    return ports[0] if ports else 10_000


def _format_aggregate_connections(
    device_name: str,
    group_ports: list[PortInfo],
    connections: dict[int, list[str]],
    client_connections: dict[int, list[str]],
    port_map: PortMap,
) -> str:
    rendered: list[str] = []
    for port in group_ports:
        port_idx = _port_index(getattr(port, "port_idx", None), getattr(port, "name", None))
        if port_idx is None:
            continue
        text = _format_connections(
            device_name,
            port_idx,
            connections,
            client_connections,
            port_map,
        )
        if text:
            rendered.append(text)
    return ", ".join([item for item in rendered if item])


def _format_aggregate_speed(group_ports: list[PortInfo]) -> str:
    speeds = {getattr(port, "speed", None) for port in group_ports}
    speeds.discard(None)
    if not speeds:
        return "-"
    if len(speeds) == 1:
        return _format_speed(next(iter(speeds)))
    return "mixed"


def _format_aggregate_poe_state(group_ports: list[PortInfo]) -> str:
    states = {_format_poe_state(port) for port in group_ports}
    if "active" in states:
        return "active"
    if "disabled" in states:
        return "disabled"
    if "capable" in states:
        return "capable"
    return "-"


def _format_aggregate_power(group_ports: list[PortInfo]) -> str:
    total = sum(getattr(port, "poe_power", 0.0) or 0.0 for port in group_ports)
    return _format_poe_power(total)



"""Shared SVG defs and theming."""

from __future__ import annotations

from dataclasses import dataclass


@dataclass(frozen=True)
class SvgTheme:
    link_standard: tuple[str, str]
    link_poe: tuple[str, str]
    node_gateway: tuple[str, str]
    node_switch: tuple[str, str]
    node_ap: tuple[str, str]
    node_client: tuple[str, str]
    node_other: tuple[str, str]


DEFAULT_THEME = SvgTheme(
    link_standard=("#16a085", "#2ecc71"),
    link_poe=("#1e88e5", "#42a5f5"),
    node_gateway=("#ffd199", "#ffb15a"),
    node_switch=("#bfe4ff", "#8ac6ff"),
    node_ap=("#c4f2d4", "#8ee3b4"),
    node_client=("#e4ccff", "#c5a4ff"),
    node_other=("#e3e3e3", "#cfcfcf"),
)


def svg_defs(prefix: str, theme: SvgTheme = DEFAULT_THEME) -> str:
    gradient_prefix = f"{prefix}-" if prefix else ""
    node_prefix = f"{prefix}-node-" if prefix else "node-"
    return (
        "<defs>"
        f'<linearGradient id="{gradient_prefix}link-standard" x1="0%" y1="0%" x2="100%" y2="0%">'
        f'<stop offset="0%" stop-color="{theme.link_standard[0]}"/>'
        f'<stop offset="100%" stop-color="{theme.link_standard[1]}"/>'
        "</linearGradient>"
        f'<linearGradient id="{gradient_prefix}link-poe" x1="0%" y1="0%" x2="100%" y2="0%">'
        f'<stop offset="0%" stop-color="{theme.link_poe[0]}"/>'
        f'<stop offset="100%" stop-color="{theme.link_poe[1]}"/>'
        "</linearGradient>"
        f'<linearGradient id="{node_prefix}gateway" x1="0%" y1="0%" x2="100%" y2="100%">'
        f'<stop offset="0%" stop-color="{theme.node_gateway[0]}"/>'
        f'<stop offset="100%" stop-color="{theme.node_gateway[1]}"/>'
        "</linearGradient>"
        f'<linearGradient id="{node_prefix}switch" x1="0%" y1="0%" x2="100%" y2="100%">'
        f'<stop offset="0%" stop-color="{theme.node_switch[0]}"/>'
        f'<stop offset="100%" stop-color="{theme.node_switch[1]}"/>'
        "</linearGradient>"
        f'<linearGradient id="{node_prefix}ap" x1="0%" y1="0%" x2="100%" y2="100%">'
        f'<stop offset="0%" stop-color="{theme.node_ap[0]}"/>'
        f'<stop offset="100%" stop-color="{theme.node_ap[1]}"/>'
        "</linearGradient>"
        f'<linearGradient id="{node_prefix}client" x1="0%" y1="0%" x2="100%" y2="100%">'
        f'<stop offset="0%" stop-color="{theme.node_client[0]}"/>'
        f'<stop offset="100%" stop-color="{theme.node_client[1]}"/>'
        "</linearGradient>"
        f'<linearGradient id="{node_prefix}other" x1="0%" y1="0%" x2="100%" y2="100%">'
        f'<stop offset="0%" stop-color="{theme.node_other[0]}"/>'
        f'<stop offset="100%" stop-color="{theme.node_other[1]}"/>'
        "</linearGradient>"
        "</defs>"
    )



import logging
from types import SimpleNamespace

from unifi_network_maps.io.debug import debug_dump_devices, device_to_dict
from unifi_network_maps.model.topology import Device


def test_device_to_dict_prefers_to_dict():
    class WithToDict:
        def to_dict(self):
            return {"name": "value"}

    assert device_to_dict(WithToDict()) == {"name": "value"}


def test_device_to_dict_uses_dunder_dict():
    class WithDict:
        def __init__(self):
            self.value = 3

    assert device_to_dict(WithDict()) == {"value": 3}


def test_device_to_dict_falls_back_to_repr():
    class NoDict:
        __slots__ = ()

        def __repr__(self):
            return "noop"

    assert device_to_dict(NoDict()) == {"repr": "noop"}


def test_debug_dump_devices_logs_gateway_and_sample(caplog):
    raw = [
        SimpleNamespace(name="Gateway", data="gw"),
        SimpleNamespace(name="Switch", data="sw"),
    ]
    normalized = [
        Device(name="Gateway", model_name="", model="", mac="aa", ip="", type="udm", lldp_info=[]),
        Device(name="Switch", model_name="", model="", mac="bb", ip="", type="usw", lldp_info=[]),
    ]

    caplog.set_level(logging.INFO)
    debug_dump_devices(raw, normalized, sample_count=1)
    assert "Gateway" in caplog.text



from __future__ import annotations

from pathlib import Path

from tests.contract_helpers import (
    assert_client_contract,
    assert_device_contract,
    load_fixture,
)
from unifi_network_maps.io.mock_data import load_mock_data
from unifi_network_maps.model.topology import (
    build_client_edges,
    build_client_port_map,
    build_device_index,
    build_edges,
    normalize_devices,
)

FIXTURES_DIR = Path(__file__).resolve().parent / "fixtures"


def _load_devices_fixture() -> list[object]:
    return load_fixture(str(FIXTURES_DIR / "unifi_devices.json"), "devices")


def _load_clients_fixture() -> list[object]:
    return load_fixture(str(FIXTURES_DIR / "unifi_clients.json"), "clients")


def test_unifi_device_fixture_contract():
    devices = _load_devices_fixture()
    assert devices
    for device in devices:
        assert_device_contract(device)


def test_unifi_client_fixture_contract():
    clients = _load_clients_fixture()
    assert clients
    for client in clients:
        assert_client_contract(client)


def test_unifi_fixture_normalization_contract():
    devices = _load_devices_fixture()
    normalized = normalize_devices(devices)
    edges = build_edges(normalized, include_ports=True, only_unifi=False)
    assert normalized
    assert edges


def test_unifi_fixture_client_edges_contract():
    devices = normalize_devices(_load_devices_fixture())
    clients = _load_clients_fixture()
    edges = build_client_edges(
        clients,
        build_device_index(devices),
        include_ports=True,
        client_mode="all",
    )
    assert any(edge.label and "Port 5" in edge.label for edge in edges)


def test_unifi_fixture_client_ports_contract():
    devices = normalize_devices(_load_devices_fixture())
    clients = _load_clients_fixture()
    client_ports = build_client_port_map(devices, clients, client_mode="all")
    rows = client_ports.get("Core Switch", [])
    assert (5, "Desk PC") in rows


def test_mock_data_fixture_contract():
    devices, clients = load_mock_data(str(Path("examples/mock_data.json")))
    normalized = normalize_devices(devices)
    edges = build_edges(normalized, include_ports=True, only_unifi=False)
    assert normalized
    assert edges
    assert clients



"""Markdown table helpers."""

from __future__ import annotations

from collections.abc import Callable, Iterable


def markdown_table_lines(
    headers: list[str],
    rows: Iterable[Iterable[str]],
    *,
    escape: Callable[[str], str] | None = None,
) -> list[str]:
    esc = escape or (lambda value: value)
    lines = [
        "| " + " | ".join(headers) + " |",
        "| " + " | ".join(["---"] * len(headers)) + " |",
    ]
    for row in rows:
        lines.append("| " + " | ".join(esc(cell) for cell in row) + " |")
    return lines



from __future__ import annotations

import os

import pytest

from tests.contract_helpers import assert_client_contract, assert_device_contract
from unifi_network_maps.adapters.config import Config
from unifi_network_maps.adapters.unifi import fetch_clients, fetch_devices

LIVE_FLAG = os.environ.get("UNIFI_CONTRACT_LIVE") == "1"


def _require_live() -> None:
    if not LIVE_FLAG:
        pytest.skip("Set UNIFI_CONTRACT_LIVE=1 to run live UniFi contract tests.")


def _load_live_config() -> Config:
    try:
        return Config.from_env()
    except ValueError as exc:
        pytest.skip(f"Missing UniFi env config: {exc}")
        raise


def _as_dict(obj: object) -> dict:
    if isinstance(obj, dict):
        return obj
    to_dict = getattr(obj, "to_dict", None)
    if callable(to_dict):
        data = to_dict()
        if isinstance(data, dict):
            return data
    if hasattr(obj, "__dict__"):
        return dict(obj.__dict__)
    return {}


def _assert_device_list_contract(devices: list[object]) -> None:
    if not devices:
        raise AssertionError("Live device list is empty")
    for device in devices[:20]:
        payload = _as_dict(device)
        if not payload:
            raise AssertionError("Device could not be coerced to dict")
        assert_device_contract(payload)


def _assert_client_list_contract(clients: list[object]) -> None:
    if not clients:
        raise AssertionError("Live client list is empty")
    for client in clients[:20]:
        payload = _as_dict(client)
        if not payload:
            raise AssertionError("Client could not be coerced to dict")
        assert_client_contract(payload)


def test_live_unifi_devices_contract():
    _require_live()
    config = _load_live_config()
    devices = list(fetch_devices(config, site=config.site, detailed=True))
    _assert_device_list_contract(devices)


def test_live_unifi_clients_contract():
    _require_live()
    config = _load_live_config()
    clients = list(fetch_clients(config, site=config.site))
    _assert_client_list_contract(clients)



"""Debug helpers for dumping device data."""

from __future__ import annotations

import json
import logging
from collections.abc import Iterable, Sequence

from ..model.topology import Device, group_devices_by_type

logger = logging.getLogger(__name__)


def device_to_dict(device: object) -> dict[str, object]:
    to_dict = getattr(device, "to_dict", None)
    if callable(to_dict):
        result = to_dict()
        if isinstance(result, dict):
            return result
        return {"repr": repr(result)}
    if hasattr(device, "__dict__"):
        return dict(device.__dict__)
    if isinstance(device, dict):
        return dict(device)
    return {"repr": repr(device)}


def debug_dump_devices(
    raw_devices: Sequence[object],
    normalized: Iterable[Device],
    *,
    sample_count: int,
) -> None:
    name_to_device: dict[str, object] = {}
    for device in raw_devices:
        name = getattr(device, "name", None)
        if name:
            name_to_device[name] = device

    groups = group_devices_by_type(normalized)
    gateways = groups.get("gateway", [])
    samples: list[str] = []
    for group in ("switch", "ap", "other"):
        for name in groups.get(group, []):
            if name not in gateways:
                samples.append(name)
            if len(samples) >= sample_count:
                break
        if len(samples) >= sample_count:
            break

    selected_names = gateways[:1] + samples
    payload = []
    for name in selected_names:
        device = name_to_device.get(name)
        if device is None:
            continue
        payload.append({"name": name, "data": device_to_dict(device)})

    logger.info("Debug dump devices: %s", json.dumps(payload, indent=2, sort_keys=True))



from unifi_network_maps.model.topology import Device, build_client_edges, build_node_type_map


def test_build_client_edges_maps_ap_mac():
    device_index = {"aa:bb:cc:dd:ee:ff": "AP One"}
    clients = [{"name": "Laptop", "ap_mac": "aa:bb:cc:dd:ee:ff", "is_wired": True}]
    edges = build_client_edges(clients, device_index)
    assert edges[0].left == "AP One"


def test_build_client_edges_uses_hostname_fallback():
    device_index = {"aa:bb:cc:dd:ee:ff": "Switch A"}
    clients = [{"hostname": "phone", "sw_mac": "aa:bb:cc:dd:ee:ff", "is_wired": True}]
    edges = build_client_edges(clients, device_index)
    assert edges[0].right == "phone"


def test_build_client_edges_skips_unknown_uplink():
    device_index = {"aa:bb:cc:dd:ee:ff": "Switch A"}
    clients = [{"name": "tablet", "sw_mac": "11:22:33:44:55:66", "is_wired": True}]
    edges = build_client_edges(clients, device_index)
    assert edges == []


def test_build_client_edges_skips_wireless_clients():
    device_index = {"aa:bb:cc:dd:ee:ff": "AP One"}
    clients = [{"name": "Laptop", "ap_mac": "aa:bb:cc:dd:ee:ff", "is_wired": False}]
    edges = build_client_edges(clients, device_index)
    assert edges == []


def test_build_client_edges_includes_wireless_when_requested():
    device_index = {"aa:bb:cc:dd:ee:ff": "AP One"}
    clients = [{"name": "Laptop", "ap_mac": "aa:bb:cc:dd:ee:ff", "is_wired": False}]
    edges = build_client_edges(clients, device_index, client_mode="wireless")
    assert edges[0].wireless is True


def test_build_client_edges_includes_uplink_port_label():
    device_index = {"aa:bb:cc:dd:ee:ff": "Switch A"}
    clients = [
        {
            "name": "Laptop",
            "sw_mac": "aa:bb:cc:dd:ee:ff",
            "is_wired": True,
            "last_uplink": {"uplink_remote_port": 3},
        }
    ]
    edges = build_client_edges(clients, device_index, include_ports=True)
    assert edges[0].label == "Switch A: Port 3 <-> Laptop"


def test_build_node_type_map_skips_wireless_clients():
    devices = [
        Device(name="Gateway", model_name="", model="", mac="aa", ip="", type="udm", lldp_info=[])
    ]
    clients = [{"name": "Phone", "is_wired": False}]
    node_types = build_node_type_map(devices, clients)
    assert "Phone" not in node_types



import re

from unifi_network_maps.model.topology import Edge
from unifi_network_maps.render.svg import render_svg_isometric


def test_render_svg_isometric_outputs_svg_root():
    output = render_svg_isometric([Edge("A", "B")], node_types={"A": "gateway", "B": "switch"})
    assert output.startswith("<svg")


def test_render_svg_isometric_includes_polygons():
    output = render_svg_isometric([Edge("A", "B")], node_types={"A": "gateway", "B": "switch"})
    assert "<path" in output


def test_render_svg_isometric_dashes_wireless_links():
    output = render_svg_isometric(
        [Edge("A", "B", wireless=True)],
        node_types={"A": "gateway", "B": "switch"},
    )
    assert 'stroke-dasharray="8 6"' in output


def _parse_points(points: str) -> list[tuple[float, float]]:
    coords = []
    for pair in points.strip().split():
        x_str, y_str = pair.split(",", 1)
        coords.append((float(x_str), float(y_str)))
    return coords


def test_isometric_label_tile_stacks_without_gap():
    output = render_svg_isometric(
        [Edge("A", "B", label="A: Port 1 <-> Port 2")],
        node_types={"A": "gateway", "B": "switch"},
    )
    polygons: list[tuple[str | None, list[tuple[float, float]]]] = []
    for match in re.finditer(r"<polygon([^>]*)>", output):
        attrs = match.group(1)
        points_match = re.search(r'points="([^"]+)"', attrs)
        if not points_match:
            continue
        cls_match = re.search(r'class="([^"]+)"', attrs)
        cls = cls_match.group(1) if cls_match else None
        polygons.append((cls, _parse_points(points_match.group(1))))

    label_index = next(i for i, (cls, _points) in enumerate(polygons) if cls == "label-tile")
    side_index = max(i for i in range(label_index) if polygons[i][0] == "label-tile-side")
    node_index = max(i for i in range(side_index) if polygons[i][0] is None)

    node_top_bottom_y = max(y for _x, y in polygons[node_index][1])
    side_bottom_y = max(y for _x, y in polygons[side_index][1])

    assert abs(side_bottom_y - node_top_bottom_y) < 0.01


def test_isometric_label_truncates_long_text():
    output = render_svg_isometric(
        [
            Edge(
                "A",
                "B",
                label="Very Long Switch Name: Port 12 <-> Port 23 Extra Extra Extra",
            )
        ],
        node_types={"A": "gateway", "B": "switch"},
    )
    tspans = re.findall(r"<tspan[^>]*>([^<]+)</tspan>", output)
    unescaped = (
        text.replace("&lt;", "<").replace("&gt;", ">").replace("&amp;", "&") for text in tspans
    )
    max_len = max(len(text) for text in unescaped)
    max_chars = int((160 * 1.5 * 0.6) / (8 * 0.6))
    assert max_len <= max_chars



import json

from unifi_network_maps.io.mock_generate import (
    MockOptions,
    generate_mock_payload,
    mock_payload_json,
)


def test_generate_mock_payload_counts():
    options = MockOptions(seed=123, switch_count=2, ap_count=3, wired_client_count=4)
    payload = generate_mock_payload(options)
    assert len(payload["devices"]) == 1 + 2 + 3
    assert len(payload["clients"]) == 4 + options.wireless_client_count


def test_generate_mock_payload_deterministic():
    options = MockOptions(seed=123, switch_count=2, ap_count=1, wired_client_count=1)
    payload_a = generate_mock_payload(options)
    payload_b = generate_mock_payload(options)
    assert payload_a == payload_b


def test_mock_payload_json_is_valid():
    options = MockOptions(seed=1)
    payload = json.loads(mock_payload_json(options))
    assert "devices" in payload
    assert "clients" in payload



import re
from pathlib import Path

import unifi_network_maps.render.svg as svg_module
from unifi_network_maps.model.topology import Edge


def test_render_svg_outputs_svg_root():
    output = svg_module.render_svg([Edge("A", "B")], node_types={"A": "gateway", "B": "switch"})
    assert output.startswith("<svg")


def test_render_svg_respects_size_override():
    output = svg_module.render_svg(
        [Edge("A", "B")],
        node_types={"A": "gateway", "B": "switch"},
        options=svg_module.SvgOptions(width=800, height=600),
    )
    assert 'width="800"' in output


def test_render_svg_escapes_edge_labels():
    output = svg_module.render_svg(
        [Edge("A", "B", label="Port 1 <-> Port 2")],
        node_types={"A": "gateway", "B": "switch"},
    )
    assert "&lt;-&gt;" in output


def test_render_svg_renders_poe_icon():
    output = svg_module.render_svg(
        [Edge("A", "B", poe=True)],
        node_types={"A": "gateway", "B": "switch"},
    )
    assert "⚡" in output


def test_render_svg_dashes_wireless_links():
    output = svg_module.render_svg(
        [Edge("A", "B", wireless=True)],
        node_types={"A": "gateway", "B": "switch"},
    )
    assert 'stroke-dasharray="6 4"' in output


def test_render_svg_adds_elbow_for_vertical_links():
    output = svg_module.render_svg(
        [Edge("Root", "Child")],
        node_types={"Root": "gateway", "Child": "switch"},
    )
    match = re.search(r'<path d="([^"]+)"', output)
    assert match is not None
    coords = [float(value) for value in re.findall(r"-?\d+(?:\.\d+)?", match.group(1))]
    x_values = {round(x, 2) for x in coords[0::2]}
    assert len(x_values) > 1


def test_render_svg_compacts_device_labels():
    output = svg_module.render_svg(
        [Edge("A", "B", label="Switch A: Port 2 <-> Switch B: Port 5")],
        node_types={"A": "gateway", "B": "switch"},
    )
    assert 'class="node-port"' in output
    assert "Switch A Port 2" in output
    assert ">5</tspan>" in output


def test_render_svg_orders_upstream_label():
    output = svg_module.render_svg(
        [Edge("Parent", "Child", label="Child: Port 1 <-> Parent: Port 2")],
        node_types={"Parent": "switch", "Child": "switch"},
    )
    assert "Parent Port 2 &lt;-&gt; Port 1" in output


def test_render_svg_moves_client_label_into_node():
    output = svg_module.render_svg(
        [Edge("Switch", "Client", label="Switch: Port 5 <-> Client")],
        node_types={"Switch": "switch", "Client": "client"},
    )
    assert 'class="node-port"' in output
    assert "Switch: Port 5" in output
    assert 'text-anchor="middle" fill="#555">Port 5' not in output


def test_render_svg_wraps_client_label():
    output = svg_module.render_svg(
        [Edge("Switch", "Client", label="Switch: Port 5 (very long uplink name)")],
        node_types={"Switch": "switch", "Client": "client"},
    )
    assert "<tspan" in output


def test_extract_port_text_non_port_prefix():
    assert svg_module._extract_port_text("eth0") is None


def test_wrap_text_splits_without_space():
    assert svg_module._wrap_text("ABCDEFGHI", max_len=4) == ["ABCD", "EFGHI"]


def test_label_metrics_empty_lines():
    assert svg_module._label_metrics([], font_size=10, padding_x=4, padding_y=3) == (8.0, 6.0)


def test_compact_edge_label_swaps_when_nodes_reversed():
    label = "B: Port 1 <-> A: Port 2"
    assert (
        svg_module._compact_edge_label(label, left_node="A", right_node="B")
        == "A Port 2 <-> Port 1"
    )


def test_render_svg_prefixes_upstream_for_port_only_label():
    output = svg_module.render_svg(
        [Edge("Switch A", "Switch B", label="Port 1 <-> Port 2")],
        node_types={"Switch A": "switch", "Switch B": "switch"},
    )
    assert "Switch A Port 1" in output


def test_render_svg_isometric_renders_label_tile():
    output = svg_module.render_svg_isometric(
        [Edge("A", "B", label="A: Port 1 <-> B: Port 2")],
        node_types={"A": "switch", "B": "switch"},
    )
    assert 'class="label-tile"' in output


def test_load_icons_missing_files_returns_empty(monkeypatch):
    monkeypatch.setattr(Path, "exists", lambda _self: False)
    assert svg_module._load_icons() == {}


def test_load_isometric_icons_missing_files_returns_empty(monkeypatch):
    monkeypatch.setattr(Path, "exists", lambda _self: False)
    assert svg_module._load_isometric_icons() == {}


def test_tree_layout_indices_cycle_returns_nodes():
    positions, _levels = svg_module._tree_layout_indices(
        [Edge("A", "B"), Edge("B", "A")],
        {"A": "switch", "B": "switch"},
    )
    assert set(positions.keys()) == {"A", "B"}


def test_tree_layout_indices_empty_returns_empty():
    positions, _levels = svg_module._tree_layout_indices([], {})
    assert positions == {}


def test_render_svg_isometric_handles_no_edges():
    output = svg_module.render_svg_isometric([], node_types={})
    assert output.startswith("<svg")


def test_render_svg_client_label_without_arrow():
    output = svg_module.render_svg(
        [Edge("Switch", "Client", label="Switch: Port 3")],
        node_types={"Switch": "switch", "Client": "client"},
    )
    assert "Switch: Port 3" in output


def test_compact_edge_label_right_port_only():
    assert svg_module._compact_edge_label("Switch <-> Port 2") == "Port 2"


def test_compact_edge_label_left_port_only():
    assert svg_module._compact_edge_label("Port 1 <-> Switch") == "Port 1"


def test_compact_edge_label_no_ports_returns_label():
    assert svg_module._compact_edge_label("A <-> B") == "A <-> B"


def test_render_svg_client_label_left_side():
    output = svg_module.render_svg(
        [Edge("Client", "Switch", label="Switch: Port 4")],
        node_types={"Switch": "switch", "Client": "client"},
    )
    assert "Switch: Port 4" in output


def test_render_svg_isometric_client_label_without_arrow():
    output = svg_module.render_svg_isometric(
        [Edge("Switch", "Client", label="Switch: Port 4")],
        node_types={"Switch": "switch", "Client": "client"},
    )
    assert "Switch: Port 4" in output


def test_render_svg_handles_missing_positions(monkeypatch):
    monkeypatch.setattr(svg_module, "_layout_nodes", lambda _e, _n, _o: ({}, 0, 0))
    output = svg_module.render_svg([Edge("A", "B")], node_types={"A": "switch", "B": "switch"})
    assert "<path" not in output


def test_render_svg_without_icons(monkeypatch):
    monkeypatch.setattr(svg_module, "_load_icons", lambda: {})
    output = svg_module.render_svg([Edge("A", "B")], node_types={"A": "switch", "B": "switch"})
    assert "<image" not in output


def test_render_svg_isometric_without_icons(monkeypatch):
    monkeypatch.setattr(svg_module, "_load_isometric_icons", lambda: {})
    output = svg_module.render_svg_isometric(
        [Edge("A", "B")], node_types={"A": "switch", "B": "switch"}
    )
    assert "<image" not in output


def test_render_svg_isometric_skips_missing_positions(monkeypatch):
    monkeypatch.setattr(svg_module, "_tree_layout_indices", lambda _e, _n: ({}, {}))
    output = svg_module.render_svg_isometric(
        [Edge("A", "B")], node_types={"A": "switch", "B": "switch"}
    )
    assert "<path" not in output


def test_render_svg_isometric_elbow_path():
    output = svg_module.render_svg_isometric(
        [Edge("Root", "B"), Edge("Root", "C")],
        node_types={"Root": "gateway", "B": "switch", "C": "switch"},
    )
    assert output.count(" L ") >= 2


def test_render_svg_isometric_poe_icon():
    output = svg_module.render_svg_isometric(
        [Edge("A", "B", poe=True)],
        node_types={"A": "switch", "B": "switch"},
    )
    assert "⚡" in output


def test_render_svg_isometric_client_left_label():
    output = svg_module.render_svg_isometric(
        [Edge("Client", "Switch", label="Switch: Port 2")],
        node_types={"Switch": "switch", "Client": "client"},
    )
    assert "Switch: Port 2" in output


def test_render_svg_isometric_port_prefixes_upstream():
    output = svg_module.render_svg_isometric(
        [Edge("Switch", "AP", label="Port 1 <-> Port 2")],
        node_types={"Switch": "switch", "AP": "ap"},
    )
    assert "Switch: Port 1" in output


def test_render_svg_isometric_defs_use_iso_node_prefix():
    output = svg_module.render_svg_isometric(
        [Edge("A", "B")], node_types={"A": "switch", "B": "switch"}
    )
    assert 'id="iso-node-switch"' in output


def test_render_svg_isometric_nodes_reference_iso_node_prefix():
    output = svg_module.render_svg_isometric(
        [Edge("A", "B")], node_types={"A": "switch", "B": "switch"}
    )
    assert 'fill="url(#iso-node-switch)"' in output



"""Edge label helpers."""

from __future__ import annotations

from collections.abc import Callable


def compose_port_label(left: str, right: str, port_map: dict[tuple[str, str], str]) -> str | None:
    left_label = port_map.get((left, right))
    right_label = port_map.get((right, left))
    if left_label and right_label:
        return f"{left}: {left_label} <-> {right}: {right_label}"
    if left_label:
        return f"{left}: {left_label} <-> {right}: ?"
    if right_label:
        return f"{left}: ? <-> {right}: {right_label}"
    return None


def order_edge_names(
    left: str,
    right: str,
    port_map: dict[tuple[str, str], str],
    rank_for_name: Callable[[str], int],
) -> tuple[str, str]:
    left_label = port_map.get((left, right))
    right_label = port_map.get((right, left))
    if left_label is None and right_label is not None:
        return (right, left)
    if left_label and right_label:
        left_rank = rank_for_name(left)
        right_rank = rank_for_name(right)
        if (left_rank, left.lower()) > (right_rank, right.lower()):
            return (right, left)
    return (left, right)



"""MkDocs-specific rendering helpers."""

from __future__ import annotations

import logging
from dataclasses import dataclass
from datetime import datetime
from zoneinfo import ZoneInfo

from ..model.topology import ClientPortMap, Device, PortMap, build_node_type_map
from .device_ports_md import render_device_port_overview
from .mermaid import render_legend, render_legend_compact, render_mermaid
from .mermaid_theme import MermaidTheme
from .templating import render_template

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class MkdocsRenderOptions:
    direction: str
    legend_style: str
    legend_scale: float
    timestamp_zone: str
    client_scope: str
    dual_theme: bool


def render_mkdocs(
    edges: list,
    devices: list[Device],
    *,
    mermaid_theme: MermaidTheme,
    port_map: PortMap,
    client_ports: ClientPortMap | None,
    options: MkdocsRenderOptions,
    dark_mermaid_theme: MermaidTheme | None = None,
) -> str:
    clients = None
    node_types = build_node_type_map(devices, clients, client_mode=options.client_scope)
    content = render_mermaid(
        edges,
        direction=options.direction,
        node_types=node_types,
        theme=mermaid_theme,
    )
    dual_theme = options.dual_theme and dark_mermaid_theme is not None
    legend_title = "Legend" if options.legend_style != "compact" else ""
    if dual_theme and dark_mermaid_theme is not None:
        dark_content = render_mermaid(
            edges,
            direction=options.direction,
            node_types=node_types,
            theme=dark_mermaid_theme,
        )
        map_block = _mkdocs_dual_mermaid_block(content, dark_content, base_class="unifi-mermaid")
        legend_block = _mkdocs_dual_legend_block(
            options.legend_style,
            mermaid_theme=mermaid_theme,
            dark_mermaid_theme=dark_mermaid_theme,
            legend_scale=options.legend_scale,
        )
        dual_style = _mkdocs_dual_theme_style()
    else:
        map_block = _mkdocs_mermaid_block(content, class_name="unifi-mermaid")
        legend_block = _mkdocs_single_legend_block(
            options.legend_style,
            mermaid_theme=mermaid_theme,
            legend_scale=options.legend_scale,
        )
        dual_style = ""
    return render_template(
        "mkdocs_document.md.j2",
        title="UniFi network",
        timestamp_line=_timestamp_line(options.timestamp_zone),
        dual_style=dual_style,
        map_block=map_block,
        legend_title=legend_title,
        legend_block=legend_block,
        device_overview=render_device_port_overview(
            devices, port_map, client_ports=client_ports
        ).rstrip()
        + "\n",
    )


def _timestamp_line(timestamp_zone: str) -> str:
    if timestamp_zone.strip().lower() in {"off", "none", "false"}:
        return ""
    try:
        zone = ZoneInfo(timestamp_zone)
    except Exception as exc:  # noqa: BLE001
        logger.warning("Invalid mkdocs timestamp zone '%s': %s", timestamp_zone, exc)
        return ""
    generated_at = datetime.now(zone).strftime("%Y-%m-%d %H:%M:%S %Z")
    return f"Generated: {generated_at}"


def _mkdocs_mermaid_block(content: str, *, class_name: str) -> str:
    return render_template(
        "mkdocs_mermaid_block.md.j2",
        class_name=class_name,
        content=content,
    )


def _mkdocs_dual_mermaid_block(
    light_content: str,
    dark_content: str,
    *,
    base_class: str,
) -> str:
    light = _mkdocs_mermaid_block(light_content, class_name=f"{base_class} {base_class}--light")
    dark = _mkdocs_mermaid_block(dark_content, class_name=f"{base_class} {base_class}--dark")
    return f"{light}\n{dark}"


def _mkdocs_single_legend_block(
    legend_style: str,
    *,
    mermaid_theme: MermaidTheme,
    legend_scale: float,
) -> str:
    if legend_style == "compact":
        return render_template(
            "mkdocs_html_block.html.j2",
            class_name="unifi-legend",
            data_unifi_legend=True,
            content=render_legend_compact(theme=mermaid_theme),
        )
    return "```mermaid\n" + render_legend(theme=mermaid_theme, legend_scale=legend_scale) + "```"


def _mkdocs_dual_legend_block(
    legend_style: str,
    *,
    mermaid_theme: MermaidTheme,
    dark_mermaid_theme: MermaidTheme,
    legend_scale: float,
) -> str:
    if legend_style == "compact":
        light = render_template(
            "mkdocs_html_block.html.j2",
            class_name="unifi-legend unifi-legend--light",
            data_unifi_legend=True,
            content=render_legend_compact(theme=mermaid_theme),
        )
        dark = render_template(
            "mkdocs_html_block.html.j2",
            class_name="unifi-legend unifi-legend--dark",
            data_unifi_legend=True,
            content=render_legend_compact(theme=dark_mermaid_theme),
        )
        return f"{light}\n{dark}"
    light = _mkdocs_mermaid_block(
        render_legend(theme=mermaid_theme, legend_scale=legend_scale),
        class_name="unifi-legend unifi-legend--light",
    )
    dark = _mkdocs_mermaid_block(
        render_legend(theme=dark_mermaid_theme, legend_scale=legend_scale),
        class_name="unifi-legend unifi-legend--dark",
    )
    return f"{light}\n{dark}"


def _mkdocs_dual_theme_style() -> str:
    return render_template("mkdocs_dual_theme_style.html.j2") + "\n"



"""CLI entry point."""

from __future__ import annotations

import argparse
import logging

from ..adapters.config import Config
from ..io.export import write_output
from ..io.mock_data import load_mock_data
from ..render.legend import render_legend_only, resolve_legend_style
from ..render.theme import resolve_themes
from .args import build_parser
from .render import render_lldp_format, render_standard_format

logger = logging.getLogger(__name__)


def _load_dotenv(env_file: str | None = None) -> None:
    try:
        from dotenv import load_dotenv
    except ImportError:
        logger.info("python-dotenv not installed; skipping .env loading")
        return
    load_dotenv(dotenv_path=env_file) if env_file else load_dotenv()


def _parse_args(argv: list[str] | None) -> argparse.Namespace:
    parser = build_parser()
    return parser.parse_args(argv)


def _load_config(args: argparse.Namespace) -> Config | None:
    try:
        _load_dotenv(args.env_file)
        return Config.from_env(env_file=args.env_file)
    except ValueError as exc:
        logging.error(str(exc))
        return None


def _resolve_site(args: argparse.Namespace, config: Config) -> str:
    return args.site or config.site


def _handle_generate_mock(args: argparse.Namespace) -> int | None:
    if not args.generate_mock:
        return None
    try:
        from ..model.mock import MockOptions, mock_payload_json
    except ImportError as exc:
        logging.error("Faker is required for --generate-mock: %s", exc)
        return 2
    options = MockOptions(
        seed=args.mock_seed,
        switch_count=max(1, args.mock_switches),
        ap_count=max(0, args.mock_aps),
        wired_client_count=max(0, args.mock_wired_clients),
        wireless_client_count=max(0, args.mock_wireless_clients),
    )
    content = mock_payload_json(options)
    write_output(content, output_path=args.generate_mock, stdout=args.stdout)
    return 0


def _load_runtime_context(
    args: argparse.Namespace,
) -> tuple[Config | None, str, list[object] | None, list[object] | None]:
    if args.mock_data:
        try:
            mock_devices, mock_clients = load_mock_data(args.mock_data)
        except Exception as exc:  # noqa: BLE001
            raise ValueError(f"Failed to load mock data: {exc}") from exc
        return None, "mock", mock_devices, mock_clients
    config = _load_config(args)
    if config is None:
        raise ValueError("Config required to run")
    site = _resolve_site(args, config)
    return config, site, None, None


def main(argv: list[str] | None = None) -> int:
    logging.basicConfig(level=logging.INFO, format="%(levelname)s %(name)s: %(message)s")
    args = _parse_args(argv)
    mock_result = _handle_generate_mock(args)
    if mock_result is not None:
        return mock_result
    try:
        config, site, mock_devices, mock_clients = _load_runtime_context(args)
    except ValueError as exc:
        logging.error(str(exc))
        return 2
    try:
        mermaid_theme, svg_theme = resolve_themes(args.theme_file)
    except Exception as exc:  # noqa: BLE001
        logging.error("Failed to load theme file: %s", exc)
        return 2

    if args.legend_only:
        legend_style = resolve_legend_style(
            format_name=args.format,
            legend_style=args.legend_style,
        )
        content = render_legend_only(
            legend_style=legend_style,
            legend_scale=args.legend_scale,
            markdown=args.markdown,
            theme=mermaid_theme,
        )
        write_output(content, output_path=args.output, stdout=args.stdout)
        return 0

    if args.format == "lldp-md":
        return render_lldp_format(
            args,
            config=config,
            site=site,
            mock_devices=mock_devices,
            mock_clients=mock_clients,
        )

    return render_standard_format(
        args,
        config=config,
        site=site,
        mock_devices=mock_devices,
        mock_clients=mock_clients,
        mermaid_theme=mermaid_theme,
        svg_theme=svg_theme,
    )


if __name__ == "__main__":
    raise SystemExit(main())



"""Render LLDP data as Markdown tables."""

from __future__ import annotations

from collections.abc import Iterable

from ..model.lldp import LLDPEntry, local_port_label
from ..model.topology import Device, build_client_port_map, build_device_index, build_port_map
from .device_ports_md import render_device_port_details
from .markdown_tables import markdown_table_lines
from .templating import render_template


def _normalize_mac(value: str) -> str:
    return value.strip().lower()


def _client_field(client: object, name: str) -> object | None:
    if isinstance(client, dict):
        return client.get(name)
    return getattr(client, name, None)


def _client_display_name(client: object) -> str | None:
    for key in ("name", "hostname", "mac"):
        value = _client_field(client, key)
        if isinstance(value, str) and value.strip():
            return value.strip()
    return None


def _client_uplink_mac(client: object) -> str | None:
    for key in ("ap_mac", "sw_mac", "uplink_mac", "uplink_device_mac", "last_uplink_mac"):
        value = _client_field(client, key)
        if isinstance(value, str) and value.strip():
            return value.strip()
    for key in ("uplink", "last_uplink"):
        nested = _client_field(client, key)
        if isinstance(nested, dict):
            value = nested.get("uplink_mac") or nested.get("uplink_device_mac")
            if isinstance(value, str) and value.strip():
                return value.strip()
    return None


def _client_uplink_port(client: object) -> int | None:
    for key in ("uplink_remote_port", "sw_port", "ap_port"):
        value = _client_field(client, key)
        if isinstance(value, int):
            return value
        if isinstance(value, str) and value.isdigit():
            return int(value)
    for key in ("uplink", "last_uplink"):
        nested = _client_field(client, key)
        if isinstance(nested, dict):
            value = nested.get("uplink_remote_port")
            if isinstance(value, int):
                return value
            if isinstance(value, str) and value.isdigit():
                return int(value)
    return None


def _client_is_wired(client: object) -> bool:
    return bool(_client_field(client, "is_wired"))


def _client_matches_mode(client: object, mode: str) -> bool:
    wired = _client_is_wired(client)
    if mode == "all":
        return True
    if mode == "wireless":
        return not wired
    return wired


def _lldp_sort_key(entry: LLDPEntry) -> tuple[int, str, str]:
    port_label = local_port_label(entry) or ""
    port_number = "".join(ch for ch in port_label if ch.isdigit())
    return (int(port_number or 0), port_label, entry.port_id)


def _port_summary(device: Device) -> str:
    ports = [port for port in device.port_table if port.port_idx is not None]
    if not ports:
        return "-"
    total_ports = len(ports)
    poe_capable = sum(1 for port in ports if port.port_poe or port.poe_enable)
    poe_active = sum(1 for port in ports if device.poe_ports.get(port.port_idx or -1))
    total_power = sum(port.poe_power or 0.0 for port in ports)
    summary = f"Total {total_ports}, PoE {poe_capable} (active {poe_active})"
    if total_power > 0:
        summary = f"{summary}, {total_power:.2f}W"
    return summary


def _poe_summary(device: Device) -> str:
    ports = [port for port in device.port_table if port.port_idx is not None]
    if not ports:
        return "-"
    poe_capable = sum(1 for port in ports if port.port_poe or port.poe_enable)
    poe_active = sum(1 for port in ports if (port.poe_power or 0.0) > 0 or port.poe_good)
    total_power = sum(port.poe_power or 0.0 for port in ports)
    summary = f"{poe_capable} capable, {poe_active} active"
    if total_power > 0:
        summary = f"{summary}, {total_power:.2f}W"
    return summary


def _uplink_summary(device: Device) -> str:
    uplink = device.uplink or device.last_uplink
    if not uplink:
        return "-"
    name = uplink.name or uplink.mac or "Unknown"
    if uplink.port is not None:
        return f"{name} (Port {uplink.port})"
    return name


def _client_summary(
    device: Device, client_rows: dict[str, list[tuple[str, str | None]]]
) -> tuple[str, str]:
    rows = client_rows.get(device.name)
    if rows is None:
        return "-", "-"
    count = len(rows)
    names = [name for name, _port in rows]
    sample = ", ".join(names[:3])
    if len(names) > 3:
        sample = f"{sample}, ..."
    return str(count), sample or "-"


def _details_table_lines(
    device: Device,
    client_rows: dict[str, list[tuple[str, str | None]]],
    client_mode: str,
) -> list[str]:
    wired_count, client_sample = _client_summary(device, client_rows)
    client_label = f"Clients ({client_mode})"
    rows = [
        ["Model", _escape_cell(device.model_name or device.type or "-")],
        ["Type", _escape_cell(device.type or "-")],
        ["IP", _escape_cell(device.ip or "-")],
        ["MAC", _escape_cell(device.mac or "-")],
        ["Firmware", _escape_cell(device.version or "-")],
        ["Uplink", _escape_cell(_uplink_summary(device))],
        ["Ports", _escape_cell(_port_summary(device))],
        ["PoE", _escape_cell(_poe_summary(device))],
        [client_label, _escape_cell(wired_count)],
        ["Client examples", _escape_cell(client_sample)],
    ]
    lines = ["### Details", ""]
    lines.extend(markdown_table_lines(["Field", "Value"], rows))
    return lines


def _lldp_rows(
    entries: Iterable[LLDPEntry],
    device_index: dict[str, str],
) -> list[list[str]]:
    rows: list[list[str]] = []
    for entry in sorted(entries, key=_lldp_sort_key):
        local_label = local_port_label(entry) or "?"
        peer_name = device_index.get(_normalize_mac(entry.chassis_id), "")
        peer_port = entry.port_id or "?"
        port_desc = entry.port_desc or ""
        rows.append(
            [
                local_label,
                peer_name or "-",
                peer_port,
                entry.chassis_id,
                port_desc or "-",
            ]
        )
    return rows


def _escape_cell(value: str) -> str:
    return value.replace("|", "\\|")


def _client_rows(
    clients: Iterable[object],
    device_index: dict[str, str],
    *,
    include_ports: bool,
    client_mode: str,
) -> dict[str, list[tuple[str, str | None]]]:
    rows_by_device: dict[str, list[tuple[str, str | None]]] = {}
    for client in clients:
        if not _client_matches_mode(client, client_mode):
            continue
        name = _client_display_name(client)
        uplink_mac = _client_uplink_mac(client)
        if not name or not uplink_mac:
            continue
        device_name = device_index.get(_normalize_mac(uplink_mac))
        if not device_name:
            continue
        port_label = None
        if include_ports:
            uplink_port = _client_uplink_port(client)
            if uplink_port is not None:
                port_label = f"Port {uplink_port}"
        rows_by_device.setdefault(device_name, []).append((name, port_label))
    return rows_by_device


def _prepare_lldp_maps(
    devices: list[Device],
    *,
    clients: Iterable[object] | None,
    include_ports: bool,
    show_clients: bool,
    client_mode: str,
) -> tuple[
    dict[tuple[str, str], str],
    dict[str, list[tuple[int, str]]] | None,
    dict[str, list[tuple[str, str | None]]],
]:
    device_index = build_device_index(devices)
    client_rows = (
        _client_rows(clients, device_index, include_ports=include_ports, client_mode=client_mode)
        if clients
        else {}
    )
    port_map: dict[tuple[str, str], str] = {}
    client_port_map = None
    if include_ports:
        port_map = build_port_map(devices, only_unifi=False)
        if clients and show_clients:
            client_port_map = build_client_port_map(devices, clients, client_mode=client_mode)
    return port_map, client_port_map, client_rows


def _render_device_lldp_section(
    device: Device,
    *,
    device_index: dict[str, str],
    port_map: dict[tuple[str, str], str],
    client_port_map: dict[str, list[tuple[int, str]]] | None,
    client_rows: dict[str, list[tuple[str, str | None]]],
    include_ports: bool,
    show_clients: bool,
    client_mode: str,
) -> str:
    details = "\n".join(_details_table_lines(device, client_rows, client_mode)).rstrip()
    ports_section = ""
    if include_ports:
        ports_section = "\n".join(
            [
                "### Ports",
                "",
                render_device_port_details(device, port_map, client_ports=client_port_map).strip(),
            ]
        ).rstrip()
    if device.lldp_info:
        lldp_section = "\n".join(
            markdown_table_lines(
                ["Local Port", "Neighbor", "Neighbor Port", "Chassis ID", "Port Description"],
                _lldp_rows(device.lldp_info, device_index),
                escape=_escape_cell,
            )
        ).rstrip()
    else:
        lldp_section = "_No LLDP neighbors._"
    clients_section = ""
    rows = client_rows.get(device.name)
    if rows and show_clients:
        if include_ports:
            clients_section = "\n".join(
                [
                    "### Clients",
                    "",
                    "\n".join(
                        markdown_table_lines(
                            ["Client", "Port"],
                            [
                                [_escape_cell(client_name), _escape_cell(port_label or "-")]
                                for client_name, port_label in rows
                            ],
                        )
                    ),
                ]
            ).rstrip()
        else:
            clients_section = "\n".join(
                ["### Clients", *[f"- {_escape_cell(name)}" for name, _ in rows]]
            ).rstrip()
    return render_template(
        "lldp_device_section.md.j2",
        device_name=device.name,
        details=details,
        ports_section=ports_section,
        lldp_section=lldp_section,
        clients_section=clients_section,
    ).rstrip()


def render_lldp_md(
    devices: list[Device],
    *,
    clients: Iterable[object] | None = None,
    include_ports: bool = False,
    show_clients: bool = False,
    client_mode: str = "wired",
) -> str:
    device_index = build_device_index(devices)
    port_map, client_port_map, client_rows = _prepare_lldp_maps(
        devices,
        clients=clients,
        include_ports=include_ports,
        show_clients=show_clients,
        client_mode=client_mode,
    )
    sections: list[str] = []
    for device in sorted(devices, key=lambda item: item.name.lower()):
        sections.append(
            _render_device_lldp_section(
                device,
                device_index=device_index,
                port_map=port_map,
                client_port_map=client_port_map,
                client_rows=client_rows,
                include_ports=include_ports,
                show_clients=show_clients,
                client_mode=client_mode,
            )
        )
    body = "\n\n".join(section for section in sections if section).rstrip()
    return (
        render_template(
            "markdown_section.md.j2",
            title="LLDP Neighbors",
            body=body,
        ).rstrip()
        + "\n"
    )



from types import SimpleNamespace

import pytest

from unifi_network_maps.model.topology import (
    Device,
    Edge,
    LLDPEntry,
    UplinkInfo,
    _aggregation_group,
    _as_bool,
    _as_float,
    _as_group_id,
    _client_display_name,
    _client_field,
    _client_uplink_mac,
    _client_uplink_port,
    _poe_ports_from_device,
    build_client_edges,
    build_edges,
    build_node_type_map,
    build_topology,
    build_tree_edges_by_topology,
    classify_device_type,
    coerce_device,
    normalize_devices,
)


class DummyDevice:
    def __init__(self, name, mac, lldp_info, port_table=None):
        self.name = name
        self.mac = mac
        self.lldp_info = lldp_info
        self.port_table = port_table or []
        self.model_name = ""
        self.ip = ""
        self.type = ""


def test_build_edges_deduplicates_links():
    dev_a = DummyDevice("Switch A", "aa:bb:cc:dd:ee:01", [LLDPEntry("aa:bb:cc:dd:ee:02", "1")])
    dev_b = DummyDevice("Switch B", "aa:bb:cc:dd:ee:02", [LLDPEntry("aa:bb:cc:dd:ee:01", "2")])
    edges = build_edges(normalize_devices([dev_a, dev_b]))
    assert len(edges) == 1


def test_build_edges_orders_deterministically():
    dev_a = DummyDevice(
        "Switch Z",
        "aa:bb:cc:dd:ee:02",
        [LLDPEntry("aa:bb:cc:dd:ee:01", "2")],
    )
    dev_b = DummyDevice(
        "Switch A",
        "aa:bb:cc:dd:ee:01",
        [LLDPEntry("aa:bb:cc:dd:ee:02", "1")],
    )
    edges = build_edges(normalize_devices([dev_a, dev_b]))
    assert [(edge.left, edge.right) for edge in edges] == [("Switch A", "Switch Z")]


def test_build_edges_includes_ports():
    dev_a = DummyDevice("Switch A", "aa:bb:cc:dd:ee:01", [LLDPEntry("aa:bb:cc:dd:ee:02", "1")])
    dev_b = DummyDevice("Switch B", "aa:bb:cc:dd:ee:02", [LLDPEntry("aa:bb:cc:dd:ee:01", "2")])
    edges = build_edges(normalize_devices([dev_a, dev_b]), include_ports=True)
    assert edges[0].label == "Switch A: 1 <-> Switch B: 2"


def test_build_edges_only_unifi_filters_unknown_neighbors():
    dev_a = DummyDevice("Switch A", "aa:bb:cc:dd:ee:01", [LLDPEntry("aa:bb:cc:dd:ee:ff", "1")])
    edges = build_edges(normalize_devices([dev_a]), only_unifi=True)
    assert edges == []


def test_build_edges_includes_unknown_neighbors_when_allowed():
    dev_a = DummyDevice("Switch A", "aa:bb:cc:dd:ee:01", [LLDPEntry("aa:bb:cc:dd:ee:ff", "1")])
    edges = build_edges(normalize_devices([dev_a]), only_unifi=False)
    assert edges[0].right == "aa:bb:cc:dd:ee:ff"


def test_build_edges_hides_mac_port_id():
    dev_switch = DummyDevice(
        "Switch A",
        "aa:bb:cc:dd:ee:01",
        [LLDPEntry("aa:bb:cc:dd:ee:02", "eth0", local_port_name="Port 2")],
    )
    dev_ap = DummyDevice(
        "AP One",
        "aa:bb:cc:dd:ee:02",
        [LLDPEntry("aa:bb:cc:dd:ee:01", "78:45:58:9F:18:38")],
    )
    edges = build_edges(normalize_devices([dev_switch, dev_ap]), include_ports=True)
    assert edges[0].label == "Switch A: Port 2 <-> AP One: ?"


def test_build_edges_port_desc_includes_number_and_name():
    dev_switch = DummyDevice(
        "Switch A",
        "aa:bb:cc:dd:ee:01",
        [
            LLDPEntry(
                "aa:bb:cc:dd:ee:02",
                "eth1",
                port_desc="uplink fiberdream",
                local_port_idx=1,
            )
        ],
        port_table=[{"port_idx": 1, "poe_enable": True}],
    )
    dev_ap = DummyDevice(
        "AP One",
        "aa:bb:cc:dd:ee:02",
        [LLDPEntry("aa:bb:cc:dd:ee:01", "eth0")],
    )
    edges = build_edges(normalize_devices([dev_switch, dev_ap]), include_ports=True)
    assert edges[0].label == "Switch A: Port 1 (uplink fiberdream) <-> AP One: Port 0"


def test_build_edges_sets_poe_when_active():
    dev_switch = DummyDevice(
        "Switch A",
        "aa:bb:cc:dd:ee:01",
        [LLDPEntry("aa:bb:cc:dd:ee:02", "eth1", local_port_idx=1)],
        port_table=[{"port_idx": 1, "poe_enable": True}],
    )
    dev_ap = DummyDevice(
        "AP One",
        "aa:bb:cc:dd:ee:02",
        [LLDPEntry("aa:bb:cc:dd:ee:01", "eth0")],
    )
    edges = build_edges(normalize_devices([dev_switch, dev_ap]))
    assert edges[0].poe is True


def test_build_edges_sets_poe_with_power():
    dev_switch = DummyDevice(
        "Switch A",
        "aa:bb:cc:dd:ee:01",
        [LLDPEntry("aa:bb:cc:dd:ee:02", "eth1", local_port_idx=1)],
        port_table=[{"port_idx": 1, "poe_power": "7.01"}],
    )
    dev_ap = DummyDevice(
        "AP One",
        "aa:bb:cc:dd:ee:02",
        [LLDPEntry("aa:bb:cc:dd:ee:01", "eth0")],
    )
    edges = build_edges(normalize_devices([dev_switch, dev_ap]))
    assert edges[0].poe is True


def test_build_edges_sets_poe_with_poe_good():
    dev_switch = DummyDevice(
        "Switch A",
        "aa:bb:cc:dd:ee:01",
        [LLDPEntry("aa:bb:cc:dd:ee:02", "eth1", local_port_idx=1)],
        port_table=[{"port_idx": 1, "poe_good": True}],
    )
    dev_ap = DummyDevice(
        "AP One",
        "aa:bb:cc:dd:ee:02",
        [LLDPEntry("aa:bb:cc:dd:ee:01", "eth0")],
    )
    edges = build_edges(normalize_devices([dev_switch, dev_ap]))
    assert edges[0].poe is True


def test_build_edges_sets_poe_with_port_poe():
    dev_switch = DummyDevice(
        "Switch A",
        "aa:bb:cc:dd:ee:01",
        [LLDPEntry("aa:bb:cc:dd:ee:02", "eth1", local_port_idx=1)],
        port_table=[{"port_idx": 1, "port_poe": True}],
    )
    dev_ap = DummyDevice(
        "AP One",
        "aa:bb:cc:dd:ee:02",
        [LLDPEntry("aa:bb:cc:dd:ee:01", "eth0")],
    )
    edges = build_edges(normalize_devices([dev_switch, dev_ap]))
    assert edges[0].poe is True


def test_coerce_device_uses_lldp_fallback():
    class DeviceWithLldp:
        name = "Device"
        model_name = ""
        mac = "aa"
        ip = ""
        type = ""
        lldp_info = None
        lldp = [LLDPEntry("bb", "1")]
        port_table = []

    device = coerce_device(DeviceWithLldp())
    assert device.lldp_info[0].chassis_id == "bb"


def test_coerce_device_uses_lldp_table_fallback():
    class DeviceWithLldpTable:
        name = "Device"
        model_name = ""
        mac = "aa"
        ip = ""
        type = ""
        lldp_info = None
        lldp = None
        lldp_table = [LLDPEntry("bb", "1")]
        port_table = []

    device = coerce_device(DeviceWithLldpTable())
    assert device.lldp_info[0].chassis_id == "bb"


def test_coerce_device_requires_name():
    class MissingName:
        name = ""
        model_name = ""
        mac = "aa"
        ip = ""
        type = ""
        lldp_info = [LLDPEntry("bb", "1")]
        port_table = []

    with pytest.raises(ValueError):
        coerce_device(MissingName())


def test_coerce_device_requires_lldp():
    class MissingLldp:
        name = "Device"
        model_name = ""
        mac = "aa"
        ip = ""
        type = ""
        lldp_info = None
        lldp = None
        port_table = []

    with pytest.raises(ValueError):
        coerce_device(MissingLldp())


@pytest.fixture()
def device_with_uplink_no_lldp():
    class MissingLldpWithUplink:
        name = "Device"
        model_name = ""
        mac = "aa"
        ip = ""
        type = ""
        lldp_info = None
        lldp = None
        uplink = {"uplink_mac": "bb", "uplink_device_name": "Gateway", "uplink_remote_port": 1}
        port_table = []

    return MissingLldpWithUplink()


def test_coerce_device_allows_uplink_when_lldp_missing(device_with_uplink_no_lldp):
    device = coerce_device(device_with_uplink_no_lldp)
    assert device.lldp_info == []


def test_build_edges_uses_uplink_fallback_fixture(device_with_uplink_no_lldp):
    gateway = Device(
        name="Gateway",
        model_name="",
        model="",
        mac="bb",
        ip="",
        type="gateway",
        lldp_info=[],
    )
    device = coerce_device(device_with_uplink_no_lldp)
    edges = build_edges([gateway, device], include_ports=True)
    assert edges[0].label == "Gateway: Port 1 <-> Device: ?"


def test_coerce_device_tracks_poe_false_when_power_invalid():
    class DeviceWithPort:
        name = "Device"
        model_name = ""
        mac = "aa"
        ip = ""
        type = ""
        lldp_info = [LLDPEntry("bb", "1")]
        port_table = [{"port_idx": 1, "poe_power": "bad"}]

    device = coerce_device(DeviceWithPort())
    assert device.poe_ports[1] is False


def test_build_tree_edges_returns_empty_without_gateways():
    edges = build_tree_edges_by_topology([Edge("A", "B")], gateways=[])
    assert edges == []


def test_coerce_device_missing_name_raises():
    device = SimpleNamespace(name=None, mac="aa", lldp_info=[])
    with pytest.raises(ValueError):
        coerce_device(device)


def test_coerce_device_missing_lldp_raises():
    device = SimpleNamespace(name="Dev", mac="aa", lldp_info=None, lldp=None)
    with pytest.raises(ValueError):
        coerce_device(device)


def test_build_edges_uses_uplink_fallback():
    gateway = Device(
        name="Gateway",
        model_name="",
        model="",
        mac="aa",
        ip="",
        type="gateway",
        lldp_info=[],
        poe_ports={1: True},
    )
    switch = Device(
        name="Switch",
        model_name="",
        model="",
        mac="bb",
        ip="",
        type="switch",
        lldp_info=[],
        uplink=UplinkInfo(mac="aa", name="Gateway", port=1),
    )
    edges = build_edges([gateway, switch], include_ports=True)
    assert edges[0].label == "Gateway: Port 1 <-> Switch: ?"


def test_poe_ports_from_device_skips_missing_port_idx():
    device = SimpleNamespace(port_table=[{"poe_enable": True}])
    assert _poe_ports_from_device(device) == {}


def test_poe_ports_from_device_reads_dict_power():
    device = SimpleNamespace(port_table=[{"port_idx": 2, "poe_power": "1.2"}])
    assert _poe_ports_from_device(device) == {2: True}


def test_poe_ports_from_device_reads_portidx_key():
    device = SimpleNamespace(port_table=[{"portIdx": 3, "poe_enable": True}])
    assert _poe_ports_from_device(device) == {3: True}


def test_as_group_id_handles_types():
    assert _as_group_id(None) is None
    assert _as_group_id(True) is None
    assert _as_group_id(5) == "5"
    assert _as_group_id(" lag1 ") == "lag1"
    assert _as_group_id(" ") is None
    assert _as_group_id(object()) is None


def test_aggregation_group_reads_dict_key():
    entry = {"lag_id": "lag5"}
    assert _aggregation_group(entry) == "lag5"


def test_aggregation_group_handles_missing_keys():
    entry = {"aggregation_group": None}
    assert _aggregation_group(entry) is None


def test_aggregation_group_reads_object_attr():
    class PortEntry:
        aggregation_group = None
        agg_id = "agg2"

    assert _aggregation_group(PortEntry()) == "agg2"


def test_client_uplink_mac_nested():
    client = {"uplink": {"uplink_mac": "aa:bb"}}
    assert _client_uplink_mac(client) == "aa:bb"


def test_client_uplink_port_nested_str():
    client = {"uplink": {"uplink_remote_port": "3"}}
    assert _client_uplink_port(client) == 3


def test_build_client_edges_skips_unwired():
    client = {"name": "Client", "is_wired": False, "sw_mac": "aa"}
    assert build_client_edges([client], {"aa": "Switch"}) == []


def test_build_edges_only_unifi_false_uses_chassis_id():
    lldp = SimpleNamespace(chassis_id="bb", local_port_idx=None, port_id="Port 1", port_desc=None)
    device = SimpleNamespace(
        name="Switch",
        model_name="",
        model="",
        mac="aa",
        ip="",
        type="switch",
        lldp_info=[lldp],
        port_table=[],
    )
    edges = build_edges([coerce_device(device)], only_unifi=False)
    assert edges[0].right == "bb"


def test_build_edges_resolves_port_idx_from_ifname():
    lldp = SimpleNamespace(
        chassis_id="bb",
        local_port_idx=None,
        local_port_name="eth1",
        port_id="Port 1",
        port_desc=None,
    )
    device = SimpleNamespace(
        name="Switch A",
        model_name="",
        model="",
        mac="aa",
        ip="",
        type="switch",
        lldp_info=[lldp],
        port_table=[{"port_idx": 2, "ifname": "eth1", "poe_enable": True}],
    )
    neighbor = SimpleNamespace(
        name="Switch B",
        model_name="",
        model="",
        mac="bb",
        ip="",
        type="switch",
        lldp_info=[],
        port_table=[],
    )
    edges = build_edges([coerce_device(device), coerce_device(neighbor)], include_ports=True)
    assert edges[0].label == "Switch A: Port 2 <-> Switch B: ?"


def test_build_tree_edges_no_gateways():
    assert build_tree_edges_by_topology([], []) == []


def test_as_bool_int_true():
    assert _as_bool(1) is True


def test_as_bool_str_truthy():
    assert _as_bool("yes") is True


def test_as_float_none_returns_zero():
    assert _as_float(None) == 0.0


def test_as_float_invalid_str_returns_zero():
    assert _as_float("nope") == 0.0


def test_as_float_int_returns_float():
    assert _as_float(2) == 2.0


def test_as_float_unknown_type_returns_zero():
    assert _as_float([]) == 0.0


def test_client_field_attribute_fallback():
    client = SimpleNamespace(name="Client")
    assert _client_field(client, "name") == "Client"


def test_client_display_name_missing_returns_none():
    client = {"name": " ", "hostname": "", "mac": ""}
    assert _client_display_name(client) is None


def test_client_uplink_port_direct_int():
    client = {"uplink_remote_port": 4}
    assert _client_uplink_port(client) == 4


def test_client_uplink_port_direct_str_digit():
    client = {"sw_port": "7"}
    assert _client_uplink_port(client) == 7


def test_client_uplink_port_nested_int():
    client = {"uplink": {"uplink_remote_port": 8}}
    assert _client_uplink_port(client) == 8


def test_client_uplink_mac_nested_empty():
    client = {"uplink": {"uplink_mac": ""}}
    assert _client_uplink_mac(client) is None


def test_build_client_edges_include_ports_without_port():
    client = {"name": "Client", "is_wired": True, "sw_mac": "aa"}
    edges = build_client_edges([client], {"aa": "Switch"}, include_ports=True)
    assert edges[0].label is None


def test_build_node_type_map_skips_unwired_clients():
    client = {"name": "Client", "is_wired": False}
    assert "Client" not in build_node_type_map([], [client])


def test_build_client_edges_missing_name_or_uplink():
    client = {"name": "", "is_wired": True}
    assert build_client_edges([client], {"aa": "Switch"}) == []


def test_build_tree_edges_gateway_not_in_adjacency():
    assert build_tree_edges_by_topology([Edge("A", "B")], ["Missing"]) == []


def test_build_client_edges_dedupes():
    clients = [
        {"name": "Client", "is_wired": True, "sw_mac": "aa"},
        {"name": "Client", "is_wired": True, "sw_mac": "aa"},
    ]
    edges = build_client_edges(clients, {"aa": "Switch"})
    assert len(edges) == 1


def test_build_node_type_map_adds_wired_client():
    client = {"name": "Client", "is_wired": True}
    node_types = build_node_type_map([], [client])
    assert node_types["Client"] == "client"


def test_classify_device_type_other():
    device = SimpleNamespace(type="camera")
    assert classify_device_type(device) == "other"


def test_build_topology_returns_edges():
    lldp = SimpleNamespace(chassis_id="bb", local_port_idx=None, port_id="Port 1", port_desc=None)
    device = SimpleNamespace(
        name="Switch",
        model_name="",
        model="",
        mac="aa",
        ip="",
        type="switch",
        lldp_info=[lldp],
        port_table=[],
    )
    result = build_topology(
        [coerce_device(device)], include_ports=False, only_unifi=False, gateways=[]
    )
    assert result.raw_edges



"""Mermaid theming helpers."""

from __future__ import annotations

from dataclasses import dataclass


@dataclass(frozen=True)
class MermaidTheme:
    node_gateway: tuple[str, str]
    node_switch: tuple[str, str]
    node_ap: tuple[str, str]
    node_client: tuple[str, str]
    node_other: tuple[str, str]
    poe_link: str
    poe_link_width: int
    poe_link_arrow: str
    standard_link: str
    standard_link_width: int
    standard_link_arrow: str
    node_text: str | None = None
    edge_label_border: str | None = None
    edge_label_border_width: int | None = None


DEFAULT_THEME = MermaidTheme(
    node_gateway=("#ffe3b3", "#d98300"),
    node_switch=("#d6ecff", "#3a7bd5"),
    node_ap=("#d7f5e7", "#27ae60"),
    node_client=("#f2e5ff", "#7f3fbf"),
    node_other=("#eeeeee", "#8f8f8f"),
    poe_link="#1e88e5",
    poe_link_width=2,
    poe_link_arrow="none",
    standard_link="#2ecc71",
    standard_link_width=2,
    standard_link_arrow="none",
    node_text=None,
    edge_label_border=None,
    edge_label_border_width=None,
)


def class_defs(theme: MermaidTheme = DEFAULT_THEME) -> list[str]:
    def node_def(name: str, fill: str, stroke: str) -> str:
        color = f",color:{theme.node_text}" if theme.node_text else ""
        return f"  classDef {name} fill:{fill},stroke:{stroke},stroke-width:1px{color};"

    return [
        node_def("node_gateway", theme.node_gateway[0], theme.node_gateway[1]),
        node_def("node_switch", theme.node_switch[0], theme.node_switch[1]),
        node_def("node_ap", theme.node_ap[0], theme.node_ap[1]),
        node_def("node_client", theme.node_client[0], theme.node_client[1]),
        node_def("node_other", theme.node_other[0], theme.node_other[1]),
        "  classDef node_legend font-size:10px;",
    ]



"""Configuration loading from environment variables."""

from __future__ import annotations

import os
from dataclasses import dataclass


def _parse_bool(value: str | None, default: bool = True) -> bool:
    if value is None:
        return default
    normalized = value.strip().lower()
    if normalized in {"1", "true", "yes", "y", "on"}:
        return True
    if normalized in {"0", "false", "no", "n", "off"}:
        return False
    return default


@dataclass(frozen=True)
class Config:
    url: str
    site: str
    user: str
    password: str
    verify_ssl: bool

    @classmethod
    def from_env(cls, *, env_file: str | None = None) -> Config:
        if env_file:
            try:
                from dotenv import load_dotenv
            except ImportError:
                raise ValueError("python-dotenv required for --env-file") from None
            load_dotenv(dotenv_path=env_file)
        url = os.environ.get("UNIFI_URL", "").strip()
        site = os.environ.get("UNIFI_SITE", "default").strip()
        user = os.environ.get("UNIFI_USER", "").strip()
        password = os.environ.get("UNIFI_PASS", "").strip()
        verify_ssl = _parse_bool(os.environ.get("UNIFI_VERIFY_SSL"), default=True)

        if not url:
            raise ValueError("UNIFI_URL is required")
        if not user:
            raise ValueError("UNIFI_USER is required")
        if not password:
            raise ValueError("UNIFI_PASS is required")

        return cls(url=url, site=site, user=user, password=password, verify_ssl=verify_ssl)
